#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from __future__ import annotations
import os, json
import pandas as pd
from .instance import Instance
from .message import Message
from .object import Object
from .behaviour import Behaviour
from .context import Context
from .model import Model
from .system import System
from ..connection import Client
from ..utils import ZendirException, printer, helper
from ..data import SimulationData


class Simulation(Context):
    """
    The Simulation class is the root object that is used to interact with the simulation.
    This can be used to create objects, behaviours, systems and messages within the simulation.
    Additionally, it can be used to track objects, query objects and save and load the state
    of the simulation. A simulation requires a client to be able to access the API. These
    clients are used to authenticate the user and ensure that the simulation is accessible.
    """

    __client: Client = None
    """Defines the client that is used to access the API."""

    __id: str = None
    """Defines the ID of the simulation. This is used to identify the simulation within the API."""

    __objects: list[Object] = []
    """Defines all objects that are created within the simulation, with the simulation root."""

    __behaviours: list[Behaviour] = []
    """Defines all behaviours that are created within the simulation, with the simulation root."""

    __systems: dict[str:System] = {}
    """Defines all systems that are created within the simulation, with the simulation root."""

    __messages: list[Message] = []
    """Defines all messages that are created within the simulation, with the simulation root."""

    __planets: dict[str:Object] = {}
    """Defines all planets that are created within the simulation, with the simulation root."""

    __time: float = 0.0
    """Defines the current time of the simulation."""

    __ticked: bool = False
    """Defines whether the simulation has been ticked or not."""

    def __init__(self, client: Client, id: str) -> None:
        """
        Initialises the simulation with the client and the ID of the simulation. If the ID is
        not provided, a new simulation will be created. If the reset flag is set to true, the simulation
        will be disposed and cleaned before initialising.

        :param client:          The client to access the API
        :type client:           Client
        :param id:              The ID of the simulation that exists
        :type id:               str
        """

        # Store a reference to the client
        self.__client: Client = client

        # If the client is are bad, throw an exception
        if not self.__client:
            raise ZendirException(
                "Failed to create a simulation due to invalid client."
            )

        # Configure the root object
        self.__id = id

        # If the ID is not valid, throw an exception
        if not helper.is_valid_guid(self.__id):
            raise ZendirException(
                "Failed to create a simulation due to invalid simulation ID."
            )

        # Reset the objects and systems
        self.__reset()

    @classmethod
    async def create(cls, client: Client) -> Simulation:
        """
        Creates a new simulation with the specified client. This will create a new simulation
        session with the API and return the simulation that has been created. If the client is
        invalid, an exception will be raised.

        :param client:          The client to access the API
        :type client:           Client

        :returns:               The simulation that has been created
        :rtype:                 Simulation
        """

        # Attempt to register a new simulation with the API
        id: str = await client.post("new", "Simulation")
        if not helper.is_valid_guid(id):
            raise ZendirException(
                "Failed to create a simulation due to invalid simulation ID."
            )

        # Create the simulation object with the ID and client
        sim: Simulation = Simulation(client, id=id)
        return sim

    @classmethod
    async def list(cls, client: Client) -> list[Simulation]:
        """
        Lists all the simulations that are currently running. This will return a list of
        simulations that are currently running and their IDs. If the client is invalid, an
        exception will be raised.

        :param client:          The client to access the API
        :type client:           Client

        :returns:               The list of simulations that are currently running
        :rtype:                 list[Simulation]
        """

        # Get the list of simulations from the API
        result = await client.get("")
        return [Simulation(client, id) for id in result]

    async def dispose(self) -> None:
        """
        Disposes of the simulation. This will dispose of the simulation and clear all objects,
        behaviours, systems and messages that have been created within the simulation. The
        simulation can no longer be used after this has been called.
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Dispose of the simulation and reset the state
        await self.__client.delete(f"{self.get_id()}", id=self.get_id())
        self.__reset()
        self.__id = None
        self.__client = None

    @classmethod
    async def dispose_all(cls, client: Client) -> None:
        """
        Disposes of all simulations that are currently running. This will dispose of all simulations
        and clear all objects, behaviours, systems and messages that have been created within the
        simulations. The simulations can no longer be used after this has been called.

        :param client:          The client to access the API
        :type client:           Client
        """

        # Dispose of all simulations
        simulations: list[Simulation] = await cls.list(client)
        for sim in simulations:
            await sim.dispose()

    def __reset(self) -> None:
        """
        Resets the simulation. This will reset the simulation and clear all objects, behaviours,
        systems and messages that have been created within the simulation. This will also reset
        the time of the simulation to zero.
        """

        # Reset the objects and systems
        self.__objects = []
        self.__behaviours = []
        self.__systems = {}
        self.__messages = []
        self.__planets = {}
        self.__time = 0.0
        self.__ticked = False

    def get_id(self) -> str:
        """
        Returns the ID of the simulation. This is used to identify the simulation within the API.

        :returns:   The ID of the simulation
        :rtype:     str
        """

        # Return the ID of the simulation, which should have been set by the API
        return self.__id

    def get_client(self) -> Client:
        """
        Returns the client that is used to access the API. This is used to authenticate the user
        and ensure that the simulation is accessible.

        :returns:   The client that is used to access the API
        :rtype:     Client
        """

        # Return the client that is used to access the API
        return self.__client

    def is_valid(self) -> bool:
        """
        Returns whether the simulation is valid or not. This will check if the simulation ID is
        valid and if the client is valid. If the simulation ID is not valid, an exception will be raised.

        :returns:   Whether the simulation is valid or not
        :rtype:     bool
        """

        # Check if the simulation ID is valid
        return (
            self.get_id() != None
            and helper.is_valid_guid(self.get_id())
            and self.__client != None
        )

    def __validate(self) -> None:
        """
        Validates the simulation. This will ensure that the simulation is valid and that the ID
        and client are set. If the simulation is not valid, an exception will be raised.
        """

        # Throw exception if the simulation is not valid
        if not self.is_valid():
            raise ZendirException(
                "Failed to call function on an invalid or deleted simulation."
            )

    async def get_function_library(self) -> System:
        """
        Returns the function library for the simulation. This will return the function library
        that is used to access the API and authenticate the user.

        :returns:   The function library for the simulation
        :rtype:     System
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Create the request to the function
        return await self.get_system("ExtensionSystem")

    def __reset_refresh_cache(self) -> None:
        """
        Ensures that the simulation requires a refresh. This will ensure that all objects, behaviours,
        systems and messages will also require a refresh.
        """

        # Ensure all sub-objects require a refresh too
        for object in self.__objects:
            object._reset_refresh_cache()
        for behaviour in self.__behaviours:
            behaviour._reset_refresh_cache()
        for system in self.__systems.values():
            system._reset_refresh_cache()
        for message in self.__messages:
            message._reset_refresh_cache()
        for planet in self.__planets.values():
            planet._reset_refresh_cache()

    def __get_registered_instance(self, id: str) -> Instance:
        """
        Attempts to get the instance with the specified ID within the simulation. This will
        search through all objects, behaviours, systems and messages to get the instance. If
        the instance is not found, None will be returned.

        :param id:      The ID of the instance to get
        :type id:       str

        :returns:       The instance with the specified ID
        :rtype:         Instance
        """

        for object in self.__objects:
            if object.get_id() == id:
                return object
            get: Instance = object.get_instance_with_id(id, True)
            if get != None:
                return get
        for behaviour in self.__behaviours:
            if behaviour.get_id() == id:
                return behaviour
            get: Instance = behaviour.get_instance_with_id(id, True)
            if get != None:
                return get
        for system in self.__systems.values():
            if system.get_id() == id:
                return system
            get: Instance = system.get_instance_with_id(id, True)
            if get != None:
                return get
        for message in self.__messages:
            if message.get_id() == id:
                return message
        return None

    async def __is_valid_instance_id(self, id: str) -> bool:
        """
        Attempts to check if the ID is a valid instance ID within the simulation. This will
        call the API to check if the ID exists and is valid. If the ID is not valid, then this
        will return False. If the ID is valid, then this will return True.

        :param id:      The ID of the instance to check
        :type id:       str

        :returns:       Whether the ID is a valid instance ID or not
        :rtype:         bool
        """

        # See if the ID already exists in the local mapping
        if not helper.is_valid_guid(id):
            raise ZendirException(
                "Failed to find a instance with an ID as the guid was incorrect."
            )

        # Create the request to the function
        result_id: str = await self.__client.post(
            f"{self.get_id()}/ivk", ["FindObjectWithID", id], id=self.get_id()
        )

        # Return whether the ID is valid or not
        return helper.is_valid_guid(result_id)

    async def add_object(self, type: str, **kwargs) -> Object:
        """
        Adds an object to the simulation with the specified type and data. This will create
        an object within the simulation and return the object that has been created. If the
        object cannot be created, an exception will be raised. This will add the object to the
        root of the simulation. An object will always have a position and orientation in the
        simulation.

        :param type:    The type of the object to create
        :type type:     str
        :param kwargs:  The optional data to set on the object
        :type kwargs:   dict

        :returns:       The object that has been created
        :rtype:         Object
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the type is valid
        type = helper.validate_type(type)

        # Create the Object ID
        object_id: str = await self.__client.post(
            f"{self.get_id()}/ivk", ["AddObject", type], id=self.get_id()
        )

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(object_id):
            raise ZendirException(f"Failed to create object of type '{type}'.")

        # Create the object and add it to the array
        object: Object = Object(self, object_id, type=type)
        self.__objects.append(object)

        # Set the data if the kwargs exist
        if len(kwargs) > 0:
            await object.set(**kwargs)

        # Print the success message
        printer.success(f"Successfully created object of type '{type}'.")

        # Return the object
        return object

    def get_objects(self, recurse: bool = True) -> list[Object]:
        """
        Returns all the objects that have been created within the simulation. This will return
        the objects as a list. If the recurse flag is set to true, all children of the objects
        will be returned as well.

        :param recurse:     Whether to return all children of the objects
        :type recurse:      bool

        :returns:           The objects that have been created within the simulation
        :rtype:             list[Object]
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Create a list of all objects.
        objects: list[Object] = self.__objects.copy()

        # If recursing, loop through all objects and get all children and then keep going
        # down the chain of children to get all objects. This might be multiple levels deep.
        if recurse:

            # Define a recursive function to get all children of an object
            def __inner_get_children(
                child: Object, children: list[Object]
            ) -> list[Object]:
                for c in child.get_children():
                    children.append(c)
                    __inner_get_children(c, children)
                return children

            # Loop through all objects and get all children
            for obj in self.__objects:
                children: list[Object] = __inner_get_children(obj, [])
                objects.extend(children)

        # Return the objects
        return objects

    def get_root_objects(self) -> list[Object]:
        """
        Returns all the root objects that have been created within the simulation. This will
        return all the objects that have been created directly within the simulation and not
        part of any other object. This will return the objects as a list.

        :returns:   The root objects that have been created within the simulation
        :rtype:     list[Object]
        """

        return self.get_objects(recurse=False)

    async def find_object_with_id(self, id: str) -> Object:
        """
        Attempts to find an object in the simulation with a specified ID. This will look through all objects
        that exist and will attempt to find one that has been created. If the object does not exist, it will
        create a Python object with the ID and, provided it exists in the simulation already, the data will
        be fetched when used.

        :param id:  The ID of the object to create
        :type id:   str

        :returns:   The object that has been found or newly created.
        :rtype:     Object
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(id):
            raise ZendirException(
                "Failed to create a object from an ID as the guid was incorrect."
            )

        # Validate if any of the current objects have the same ID
        for obj in self.__objects:
            if obj.get_id() == id:
                return obj

            # Check children
            obj: Object = await obj.find_child_with_id(id, recurse=True)
            if obj != None:
                return obj

        # Attempt to find the object with the ID from the API
        if not await self.__is_valid_instance_id(id):
            return None

        # Otherwise, assume it is a root object and create the Object instance
        object: Object = Object(self, id)
        self.__objects.append(object)

        # Print the success message
        printer.success(f"Successfully created object with ID '{id}'.")

        # Return the object that has been found
        return object

    async def add_behaviour(self, type: str, **kwargs) -> Behaviour:
        """
        Adds a behaviour to the simulation with the specified type and data. This will create
        a behaviour within the simulation and return the behaviour that has been created. If the
        behaviour cannot be created, an exception will be raised. This will add the behaviour to
        the root of the simulation. A behaviour is a type of object that can be attached to other
        objects to provide additional functionality, but has no physical representation.

        :param type:    The type of the behaviour to create
        :type type:     str
        :param kwargs:  The optional data to set on the behaviour
        :type kwargs:   dict

        :returns:       The behaviour that has been created
        :rtype:         Behaviour
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the type is valid
        type = helper.validate_type(type)

        # Create the behaviour ID
        behaviour_id: str = await self.__client.post(
            f"{self.get_id()}/ivk", ["AddObject", type], id=self.get_id()
        )

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(behaviour_id):
            raise ZendirException(f"Failed to create behaviour of type '{type}'.")

        # Create the behaviour and add it to the array
        behaviour: Behaviour = Behaviour(self, behaviour_id, type=type, parent=None)
        self.__behaviours.append(behaviour)

        # Set the data if the kwargs exist
        if len(kwargs) > 0:
            await behaviour.set(**kwargs)

        # Print the success message
        printer.success(f"Successfully created behaviour of type '{type}'.")

        # Return the behaviour
        return behaviour

    def get_behaviours(self, recurse: bool = True) -> list[Behaviour]:
        """
        Returns all the behaviours that have been created within the simulation. This will return
        the behaviours as a list. If the recurse flag is set to true, all child behaviours of the objects
        will be returned as well.

        :param recurse:     Whether to return all behaviours of the objects
        :type recurse:      bool

        :returns:           The behaviours that have been created within the simulation
        :rtype:             list[Behaviour]
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Create a list of all behaviours
        behaviours: list[Behaviour] = self.__behaviours.copy()

        # If recursing, loop through all objects and get all behaviours
        if recurse:
            for obj in self.get_objects(True):

                # Get the children of the object
                child_behaviours: list[Behaviour] = obj.get_behaviours()
                behaviours.extend(child_behaviours)

        # Return the behaviours
        return behaviours

    def get_root_behaviours(self) -> list[Behaviour]:
        """
        Returns all the root behaviours that have been created within the simulation. This will
        return all the behaviours that have been created directly within the simulation and not
        part of any other object. This will return the behaviours as a list.

        :returns:   The root behaviours that have been created within the simulation
        :rtype:     list[Behaviour]
        """

        return self.get_behaviours(recurse=False)

    async def find_behaviour_with_id(self, id: str) -> Behaviour:
        """
        Attempts to find a behaviour in the simulation with a specified ID. This will look through all behaviours
        that exist and will attempt to find one that has been created.

        :param id:  The ID of the behaviour to get
        :type id:   str

        :returns:   The behaviour that has been found
        :rtype:     Behaviour
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(id):
            raise ZendirException("Failed to find behaviour with ID.")

        # Validate if any of the current behaviours have the same ID
        for behaviour in self.__behaviours:
            if behaviour.get_id() == id:
                return behaviour

        # Fetch the children
        for child in self.get_objects(True):

            # Check if the behaviour is in the children
            for behaviour in child.get_behaviours():
                if behaviour.get_id() == id:
                    return behaviour

        # Attempt to find the object with the ID from the API
        if not await self.__is_valid_instance_id(id):
            return None

        # Otherwise, assume it is a root behaviour and create the Behaviour instance
        behaviour: Behaviour = Behaviour(self, id)
        self.__behaviours.append(behaviour)

        # Print the success message
        printer.success(f"Successfully created behaviour with ID '{id}'.")

        # Return the behaviour that has been found
        return behaviour

    async def get_system(self, type: str, **kwargs) -> System:
        """
        Attempts to get the system with the specified type within the simulation. If the system
        does not exist, it will be created with the specified type and data. If the system cannot
        be created, an exception will be raised. This will add the system to the root of the simulation.

        :param type:    The type of the system to get or create
        :type type:     str
        :param kwargs:  The optional data to set on the system
        :type kwargs:   dict

        :returns:       The system with the specified type
        :rtype:         System
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the type is valid
        type = helper.validate_type(type)

        # Check if the system exists and return it
        if type in self.__systems:
            system: System = self.__systems[type]

            # Update the values if there are any
            if len(kwargs) > 0:
                await system.set(**kwargs)
            return system

        # Attempt to find the object of type
        id: str = await self.__client.post(
            f"{self.get_id()}/ivk", ["FindObjectWithType", type], id=self.get_id()
        )
        if not helper.is_valid_guid(id):

            # Attempt to create the system
            id = await self.__client.post(
                f"{self.get_id()}/ivk", ["AddObject", type], id=self.get_id()
            )

            # If the ID is not valid, raise an exception
            if not helper.is_valid_guid(id):
                raise ZendirException(f"Failed to create system of type '{type}'.")

        # Create the system object from the ID and update any kwargs if there are any
        system: System = System(self, id, type=type)
        if len(kwargs) > 0:
            await system.set(**kwargs)
        self.__systems[type] = system

        # Print the success message
        printer.success(f"Successfully created system of type '{type}'.")
        return system

    def get_systems(self) -> list[System]:
        """
        Returns all the systems that have been created within the simulation. This will return
        all the systems that have been created within the simulation as a list.

        :returns:   The systems that have been created within the simulation
        :rtype:     list[System]
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Return the systems that have been created within the simulation
        return list(self.__systems.values())

    def get_models(self) -> list[Model]:
        """
        Returns all the models that have been created within the simulation. This will return
        all the models that have been created within the simulation as a list.

        :returns:   The models that have been created within the simulation
        :rtype:     list[Model]
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Fetch all objects
        objects: list[Object] = self.get_objects(True)
        models: list[Model] = []

        # Loop through all objects and get all models
        for obj in objects:
            child_models: list[Model] = obj.get_models()
            models.extend(child_models)

        # Return the models that have been created within the simulation
        return models

    async def add_message(self, type: str, **kwargs) -> Message:
        """
        Creates a message within the simulation with the specified type and data. This will create
        a message within the simulation and return the message that has been created. If the message
        cannot be created, an exception will be raised. This will add the message to the root of the
        simulation. A message is a type of object that can be used to communicate between objects and
        systems within the simulation.

        :param type:    The type of the message to create
        :type type:     str
        :param kwargs:  The optional data to set on the message
        :type kwargs:   dict

        :returns:       The message that has been created
        :rtype:         Message
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the type is valid
        type = helper.validate_type(type)

        # Create the message ID
        message_id: str = await self.__client.post(
            f"{self.get_id()}/ivk", ["AddObject", type], id=self.get_id()
        )

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(message_id):
            raise ZendirException(f"Failed to create message of type '{type}'.")

        # Create the message and add it to the array
        message: Message = Message(self, message_id, type=type)
        self.__messages.append(message)

        # Set the data if the kwargs exist
        if len(kwargs) > 0:
            await message.set(**kwargs)

        # Print the success message
        printer.success(f"Successfully created message of type '{type}'.")

        # Return the message
        return message

    def get_messages(self, recurse: bool = True) -> list[Message]:
        """
        Returns all the messages that have been created within the simulation. This will return
        the messages as a list. If the recurse flag is set to true, all child messages of the instances
        will be returned as well.

        :param recurse:     Whether to return all messages of the instances
        :type recurse:      bool

        :returns:           The messages that have been created within the simulation
        :rtype:             list[Message]
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Create a list of all messages
        messages: list[Message] = self.__messages.copy()

        # If recursing, loop through all objects and get all messages
        if recurse:

            # Loop through all objects and get all messages
            for obj in self.get_objects(True):
                child_messages: list[Message] = obj._Object__messages.values()
                messages.extend(child_messages)

            # Loop through all behaviours and get all messages
            for behaviour in self.get_behaviours(True):
                child_messages: list[Message] = behaviour._Behaviour__messages.values()
                messages.extend(child_messages)

            # Loop through all systems and get all messages
            for system in self.get_systems():
                child_messages: list[Message] = system._System__messages.values()
                messages.extend(child_messages)

            # Loop through all models and get all messages
            for model in self.get_models():
                child_messages: list[Message] = model._Model__messages.values()
                messages.extend(child_messages)

        # Return the messages
        return messages

    def get_root_messages(self) -> list[Message]:
        """
        Returns all the root messages that have been created within the simulation. This will
        return all the messages that have been created directly within the simulation and not
        part of any other object. This will return the messages as a list.

        :returns:   The root messages that have been created within the simulation
        :rtype:     list[Message]
        """

        return self.get_messages(recurse=False)

    async def get_planet(self, name: str) -> Object:
        """
        Returns the planet with the specified name within the simulation. This will return the
        planet object that has been created within the simulation. If the planet does not exist,
        it will be created and returned. If the planet cannot be created, None will be returned.

        :param name:    The name of the planet to get or create
        :type name:     str

        :returns:       The planet with the specified name
        :rtype:         Object
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the planet already exists
        name = name.strip().lower()
        if name in self.__planets:
            return self.__planets[name]

        # Fetch the solar system
        system: System = await self.get_system("SolarSystem")

        # Grab the planet by invoking the method
        id: str = await system.invoke("GetBody", name)
        if not helper.is_valid_guid(id):
            raise ZendirException(f"Failed to create planet with name '{name}'.")

        # Construct the object
        object: Object = Object(self, id, "CelestialBody")
        self.__planets[name] = object
        self.__objects.append(object)

        # Return the object
        return object

    async def find_message_with_id(self, id: str) -> Message:
        """
        Attempts to find a message in the simulation with a specified ID. This will look through all messages
        that exist and will attempt to find one that has been created. If the message does not exist, it will
        create a Python message with the ID and, provided it exists in the simulation already, the data will
        be fetched when used.

        :param id:  The ID of the message to create
        :type id:   str

        :returns:   The message that has been found or newly created.
        :rtype:     Message
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(id):
            raise ZendirException(
                "Failed to create a message from an ID as the guid was incorrect."
            )

        # Validate if any of the current messages have the same ID
        for message in self.__messages:
            if message.get_id() == id:
                return message

        # Loop through all objects and check for messages
        for message in self.get_messages(recurse=True):
            if message.get_id() == id:
                return message

        # Attempt to find the message with the ID from the API
        if not await self.__is_valid_instance_id(id):
            return None

        # Otherwise, assume it is a root message and create the Message instance
        message: Message = Message(self, id)
        self.__messages.append(message)

        # Print the success message
        printer.success(f"Successfully created message with ID '{id}'.")

        # Return the message that has been found
        return message

    async def find_instance_with_type(self, type: str) -> Instance:
        """
        Searches for an instance of the specified type within the simulation. If the instance
        does not exist, it will be created and returned. If the instance cannot be created, None
        will be returned. This will search through all objects, behaviours, systems and messages
        within the simulation to find the instance.

        :param type:    The type of the instance to find
        :type type:     str

        :returns:       The instance of the specified type
        :rtype:         Instance
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the type is valid
        type = helper.validate_type(type)

        # Create the request to the function
        id: str = await self.__client.post(
            f"{self.get_id()}/ivk", ["FindObjectWithType", type], id=self.get_id()
        )

        # If the result is not a valid GUID, return None
        if not helper.is_valid_guid(id):
            return None

        # Otherwise, check if the instance is already in the list
        instance: Instance = self.__get_registered_instance(id)
        if instance != None:
            return instance

        # Otherwise, create the instance and return it
        return Instance(self, id, type)

    async def find_instances_with_type(self, type: str) -> list[Instance]:
        """
        Finds all instances of the specified type within the simulation. This will search through
        all objects, behaviours, systems and messages within the simulation to find the instances.
        If the instances do not exist, an empty list will be returned.

        :param type:    The type of the instances to find
        :type type:     str

        :returns:       The instances of the specified type
        :rtype:         list[Instance]
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Check if the type is valid
        type = helper.validate_type(type)

        # Create the request to the function
        result = await self.__client.post(
            f"{self.get_id()}/ivk", ["FindObjectsWithType", type], id=self.get_id()
        )

        # If the result is not a list or is empty, return a missing list
        if not isinstance(result, list) or len(result) == 0:
            return []

        # Loop through the list
        instances: list = []
        for i in range(len(result)):
            instance: Instance = self.__get_registered_instance(result[i])
            if instance != None:
                instances.append(instance)
            else:
                instances.append(Instance(self, result[i], type))

        # Return the list
        return instances

    async def find_instance_with_id(self, id: str) -> Instance:
        """
        Searches for an instance of the specified ID within the simulation. If the instance
        does not exist, it will not be created. If the instance cannot be created, None
        will be returned. This will search through all objects, behaviours, systems and messages
        within the simulation to find the instance.

        :param id:    The ID of the instance to find
        :type id:     str

        :returns:       The instance of the specified ID
        :rtype:         Instance
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # See if the ID already exists in the local mapping
        if not helper.is_valid_guid(id):
            raise ZendirException(
                "Failed to find a instance with an ID as the guid was incorrect."
            )

        # If the ID is already in the list, return it
        instance: Instance = self.__get_registered_instance(id)
        if instance != None:
            return instance

        # Create the request to the function
        if not await self.__is_valid_instance_id(id):
            return None

        # Otherwise, create the instance and return it
        return Instance(self, id)

    async def __load_cache(self) -> None:
        """
        Loads the cache for the simulation. This will load all objects, behaviours, systems and messages
        that have been created within the simulation. This will create all the appropriate instances and
        objects in the correct structure.
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Get the extension system
        system: System = await self.get_function_library()

        # Get the structure of the simulation
        structure: dict = await system.invoke("GetSimulationStructure")
        if structure is None:
            raise ZendirException("Failed to load the simulation structure.")

        # Set the time of the simulation
        self.__time = float(structure.get("Time", 0.0))
        self.__ticked = self.__time > 0.0

        # Defines a function to register a message with the instance
        def __register_msg(instance: Instance, msg_data: dict) -> None:
            msg: Message = Message(self, msg_data["ID"], type=msg_data["Type"])
            name: str = msg_data.get("Name", "")
            if type(instance) is System:
                if name not in instance._System__messages:
                    instance._System__messages[name] = msg
            elif type(instance) is Behaviour:
                if name not in instance._Behaviour__messages:
                    instance._Behaviour__messages[name] = msg
            elif type(instance) is Object:
                if name not in instance._Object__messages:
                    instance._Object__messages[name] = msg
                    instance._Object__instances[msg.get_id()] = msg
            elif type(instance) is Model:
                if name not in instance._Model__messages:
                    instance._Model__messages[name] = msg

        # Defines a function to register a behaviour
        def __register_behaviour(behaviour: Behaviour, behaviour_data: dict) -> None:

            # Loop through all messages and register them
            for msg_data in behaviour_data.get("Messages", []):
                __register_msg(behaviour, msg_data)

        # Defines a function to register a model
        def __register_model(model: Model, model_data: dict) -> None:

            # Loop through all messages and register them
            for msg_data in model_data.get("Messages", []):
                __register_msg(model, msg_data)

        # Defines a function to register an object
        def __register_object(object: Object, object_data: dict) -> None:

            # Loop through all messages and register them
            for msg_data in object_data.get("Messages", []):
                __register_msg(object, msg_data)

            # Loop through all models and register them
            for model_data in object_data.get("Models", []):
                model_type: str = model_data["Type"]
                model: Model = None
                if model_type in object._Object__models:
                    model = object._Object__models[model_type]
                else:
                    model: Model = Model(
                        self, model_data["ID"], type=model_type, parent=object
                    )
                    object._Object__models[model_type] = model
                    object._Object__instances[model.get_id()] = model
                __register_model(model, model_data)

            # Loop through all behaviours and register them
            for behaviour_data in object_data.get("Behaviours", []):
                behaviour: Behaviour = None
                for b in object._Object__behaviours:
                    if b.get_id() == behaviour_data["ID"]:
                        behaviour = b
                        break
                if behaviour is None:
                    behaviour: Behaviour = Behaviour(
                        self,
                        id=behaviour_data["ID"],
                        type=behaviour_data["Type"],
                        parent=object,
                    )
                    object._Object__behaviours.append(behaviour)
                    object._Object__instances[behaviour.get_id()] = behaviour
                __register_behaviour(behaviour, behaviour_data)

            # Loop through all children and register them
            for child_data in object_data.get("Children", []):
                child: Object = None
                for c in object._Object__children:
                    if c.get_id() == child_data["ID"]:
                        child = c
                        break
                if child is None:
                    child: Object = Object(
                        self,
                        id=child_data["ID"],
                        type=child_data["Type"],
                        parent=object,
                    )
                    object._Object__children.append(child)
                    object._Object__instances[child.get_id()] = child
                __register_object(child, child_data)

        # Loop through all systems and create them
        for system_data in structure.get("Systems", []):

            # If the system already exists, get it, otherwise create it
            system: System = None
            if system_data["Type"] in self.__systems:
                system = self.__systems[system_data["Type"]]
            else:
                system: System = System(
                    self, id=system_data["ID"], type=system_data["Type"]
                )
                self.__systems[system._Instance__type] = system

            # Register the messages in each system
            messages: list = system_data.get("Messages", [])
            for msg_data in messages:
                __register_msg(system, msg_data)

        # Loop through all behaviours and create them
        for behaviour_data in structure.get("Behaviours", []):

            # If the behaviour already exists, get it, otherwise create it
            behaviour: Behaviour = None
            for b in self.__behaviours:
                if b.get_id() == behaviour_data["ID"]:
                    behaviour = b
                    break

            # If the behaviour was not found, create it
            if behaviour is None:
                behaviour: Behaviour = Behaviour(
                    self, id=behaviour_data["ID"], type=behaviour_data["Type"]
                )
                self.__behaviours.append(behaviour)

            # Register the behaviour data
            __register_behaviour(behaviour, behaviour_data)

        # Loop through all objects and create them
        for object_data in structure.get("Objects", []):

            # If the object already exists, get it, otherwise create it
            object: Object = None
            for o in self.__objects:
                if o.get_id() == object_data["ID"]:
                    object = o
                    break

            # If the object was not found, create it
            if object is None:
                object: Object = Object(
                    self, id=object_data["ID"], type=object_data["Type"]
                )
                self.__objects.append(object)

            # Register the object data
            __register_object(object, object_data)

    async def get_state(self) -> dict:
        """
        Returns the state of the simulation. This will fetch the state from the API and return
        the state as a dictionary.

        :returns:   The state of the simulation
        :rtype:     dict
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Get the extension system
        system: System = await self.get_function_library()

        # Store the total data and the current page being called
        data: str = ""
        page_count: int = 1
        page: int = 0

        # Loop through all pages (which is at least 1)
        while page < page_count:

            # Invoke the query object on the API, with the current page
            page_data: dict = await system.invoke("GetState", page)
            if page_data == None:
                return None

            # If the first page, then store the page data including the metadata
            if page == 0:
                data = page_data["Data"]

            # Otherwise, simply extend the data from the next pages
            else:
                data += page_data["Data"]

            # Update the page count
            # Handle the case where the page count is not present in the page data
            if "Count" not in page_data:
                raise ZendirException(
                    "No data can be retrieved for this query. Make sure the data was subscribed to."
                )
            page_count = page_data["Count"]
            page += 1

        # Finally, we need to deserialize the data
        try:
            return json.loads(data)
        except json.JSONDecodeError as e:
            raise ZendirException("Failed to decode the state data as JSON.")

    async def save_state(self, path: str) -> None:
        """
        Saves the state of the simulation to the specified path. This will save the state of the
        simulation to the path as a JSON file. If the path does not exist, an exception will be raised.

        :param path:    The path to save the state of the simulation to
        :type path:     str
        """

        # Get the state of the simulation as a raw JSON
        state: dict = await self.get_state()

        # Create the directory if it does not exist
        directory: str = os.path.dirname(path)
        if not os.path.exists(directory):
            try:
                os.makedirs(directory)
            except:
                pass

        # Save the state to the path
        with open(path, "w") as file:
            json.dump(state, file)

    async def set_state(self, state: dict) -> bool:
        """
        Sets the state of the simulation to the specified state from a JSON object. This
        will also reset and load all objects, behaviours, systems and messages within the
        simulation correctly, storing them into the Python module.

        :param state:       The state to set the simulation to
        :type state:        dict

        :returns:       Whether the state was set successfully
        :rtype:         bool
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Clear the current state
        self.__reset()

        # Get the extension system
        system: System = await self.get_function_library()

        # Start by converting the state to a JSON string
        state_json: str = json.dumps(state)

        # Break it up into chunks of a maximum amount of characters
        chunks: list[str] = []
        CHUNK_SIZE: int = self.__client.get_chunk_size()
        for i in range(0, len(state_json), CHUNK_SIZE):
            chunks.append(state_json[i : i + CHUNK_SIZE])

        # Now, we need to load the chunks individually
        for index, chunk in enumerate(chunks):
            success: bool = await system.invoke("SetState", chunk, index, len(chunks))
            if not success:
                return False

        # Next, load the cache
        await self.__load_cache()

        # Ensure the refresh is required
        self.__reset_refresh_cache()

        # Return the success
        return True

    async def load_state(self, path: str) -> bool:
        """
        Loads the state of the simulation from the specified path. This will load the state of the
        simulation from the path as a JSON file and return whether the state was loaded successfully.
        If the path does not exist, an exception will be raised. This will also reset and load all objects,
        behaviours, systems and messages within the simulation correctly, storing them into the Python module.

        :param path:        The path to load the state of the simulation from
        :type path:         str

        :returns:       Whether the state was loaded successfully
        :rtype:         bool
        """

        # Check if the path exists
        if not os.path.exists(path):
            raise ZendirException(
                f"Failed to load state from path '{path}' as it does not exist."
            )

        # Load the state from the path
        with open(path, "r") as file:
            state: dict = json.load(file)
            return await self.set_state(state)

    async def reload_data(self) -> None:
        """
        Reloads the data of the simulation. This will reload all objects, behaviours, systems and messages
        within the simulation correctly, storing them into the Python module. This is useful if the simulation
        has been modified externally and the data needs to be refreshed.
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Clear the current state
        self.__reset()

        # Load the cache
        await self.__load_cache()

        # Ensure the refresh is required
        self.__reset_refresh_cache()

    async def get_time(self) -> float:
        """
        Returns the current time of the simulation. This will fetch the time from the API
        and return the time as a floating point number.

        :returns:   The current time of the simulation in seconds
        :rtype:     float
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # If the time is not zero, return the time
        if self.__time > 0:
            return self.__time

        # Get the extension system and grab the time
        system: System = await self.get_function_library()
        self.__time = float(await system.invoke("GetSimulationTime"))

        # Return the time
        return self.__time

    async def tick(self, step: float = 0.1) -> None:
        """
        Ticks the simulation by the specified amount of time. This will invoke the tick function
        on the simulation and update the time by the specified amount. If the step is not provided,
        the default step of 0.1 seconds will be used.

        :param step:    The amount of time to tick the simulation by in seconds
        :type step:     float
        """

        # Tick the simulation by the specified amount of time
        await self.tick_duration(step, step)

    async def tick_duration(self, time: float, step: float = 0.1) -> None:
        """
        Ticks the simulation by the specified amount of time. This will invoke the tick function
        on the simulation and update the time by the specified amount. If the step is not provided,
        the default step of 0.1 seconds will be used. If the step is too small, it will do the final
        step with the remaining time.

        :param time:    The totalamount of time to tick the simulation by in seconds
        :type time:     float
        :param step:    The time-step of the tick, used for physics calculations in seconds
        :type step:     float
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # If the time is zero, throw an exception
        if time < 1e-6:
            raise ZendirException(
                "Failed to tick simulation as the time is zero or negative."
            )

        # If the step is zero, throw an exception
        if step < 1e-6:
            raise ZendirException(
                "Failed to tick simulation as the step is zero or negative."
            )

        # Get the extension system
        system: System = await self.get_function_library()

        # If the simulation has not been ticked yet, call a tick with 0.0 to initialise the tracking data
        if not self.__ticked:
            await system.invoke("InitializeSimulation")
            self.__ticked = True

        # Calculate the number of steps to take
        iterations = int(time / step)

        # Determine the final time to tick to (roundede to 1microsecond precision)
        final_time: float = round(time + self.__time, 6)

        # While there are steps remaining, tick
        while iterations > 0:

            # Tick the iterations and get the amount of iterations completed
            result = await system.invoke("TickIterations", iterations, step)
            iterations -= result

            # Update the time
            self.__time += result * step

        # Check if there is any remaining time to tick
        if abs(self.__time - final_time) > 1e-6:
            await system.invoke("TickIterations", 1, final_time - self.__time)
            self.__time = final_time

        # Ensure the refresh is required
        self.__reset_refresh_cache()

    async def get_tracking_interval(self) -> float:
        """
        Returns the current tracking interval for all the objects that have been tracked. This will
        return the interval in simulation seconds. The default value is 10.0 seconds.

        :returns:   The interval to track the objects by in seconds
        :rtype:     float
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Get the tracking system
        system: System = await self.get_system("TrackingSystem")

        # Invoke the get track interval
        return await system.get("Interval")

    async def set_tracking_interval(self, interval: float) -> None:
        """
        Updates the tracking interval for all the objects that have been tracked. This will
        be an interval in simulation seconds and will be the same across all objects. The
        default value is 10.0 seconds.

        :param interval:    The interval to track the objects by in seconds
        :type interval:     float
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # If the interval is not positive, throw an exception
        if interval < 1e-6:
            raise ZendirException(
                "Failed to set tracking interval as the interval must be positive."
            )

        # Get the tracking system
        system: System = await self.get_system("TrackingSystem")

        # Invoke the set track interval
        await system.set(Interval=interval)

    async def track_object(self, instance: Instance, isAdvanced: bool = False) -> None:
        """
        Starts tracking the object within the simulation. This will start tracking the object
        which will store the object's data and state every specified tick in a local database.
        This can be used to retrieve the object's data at a later time.

        :param instance:    The instance to track within the simulation
        :type instance:     Instance
        :param isAdvanced:  Whether the advanced variables are also being tracked
        :type isAdvanced:   bool
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Get the tracking system
        system: System = await self.get_system("TrackingSystem")

        # Get the instance ID, based on whether it is an instance or a string
        id: str = instance.get_id() if isinstance(instance, Instance) else instance

        # If the ID is not a valid GUID, raise an exception
        if not helper.is_valid_guid(id):
            raise ZendirException("Failed to track object as the ID was not valid.")

        # Invoke the track object on the API, with the ID and whether it is advanced
        await system.invoke("TrackObject", id, isAdvanced)

    async def query_object(self, instance: Instance) -> SimulationData:
        """
        Queries the object within the simulation. This will query the object and return the data
        that has been stored for the object. This will return the data as a data frame. If there
        is no data for the object, or if the object has not yet been tracked, an empty data frame
        will be returned. Due to the API having a maximum limit of data, this will also fetch the
        data in pages before pre-compiling the data into a single data frame.

        :param instance:    The instance to query within the simulation
        :type instance:     Instance

        :returns:           The data that has been stored for the object
        :rtype:             DataFrame
        """

        # Throw exception if the simulation is not valid
        self.__validate()

        # Get the tracking system
        system: System = await self.get_system("TrackingSystem")

        # Store the total data and the current page being called
        data: dict = {}
        page_count: int = 1
        page: int = 0

        # Get the instance ID, based on whether it is an instance or a string
        id: str = instance.get_id() if isinstance(instance, Instance) else instance

        # If the ID is not a valid GUID, raise an exception
        if not helper.is_valid_guid(id):
            raise ZendirException("Failed to query object as the ID was not valid.")

        # Loop through all pages (which is at least 1)
        while page < page_count:

            # Invoke the query object on the API, with the current page
            page_data: dict = await system.invoke("ExportToAPI", id, page)
            if page_data == None:
                return None

            # If the first page, then store the page data including the metadata
            if page == 0:
                data = page_data

            # Otherwise, simply extend the data from the next pages
            else:
                data["Data"].extend(page_data["Data"])

            # Update the page count
            # Handle the case where the page count is not present in the page data
            if "Count" not in page_data:
                raise ZendirException(
                    "No data can be retrieved for this query. Make sure the data was subscribed to."
                )
            page_count = page_data["Count"]
            page += 1

        # Remove the page information as it is not needed
        del data["Page"]
        del data["Count"]

        # Create and return the data frame
        return SimulationData(data)

    async def query_dataframe(self, instance: Instance) -> pd.DataFrame:
        """
        Queries the object within the simulation. This will query the object and return the data
        that has been stored for the object. This will return the data as a data frame. If there
        is no data for the object, or if the object has not yet been tracked, an empty data frame
        will be returned.

        :param instance:    The instance to query within the simulation
        :type instance:     Instance

        :returns:           The data that has been stored for the object
        :rtype:             DataFrame
        """

        # Create and return the data frame
        return (await self.query_object(instance=instance)).to_dataframe()



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

"""
This module contains helper functions for printing lines to the console.
By default, any calls made to these functions will not print any data.
If the verbosity is enabled, then data will be printed to the console.
"""

import datetime
import os


# Defines ANSI escape sequences for colors
__RESET = "\033[0m"
__RED = "\033[31m"
__GREEN = "\033[32m"
__YELLOW = "\033[33m"
__BLUE = "\033[34m"
__PURPLE = "\033[35m"
__CYAN = "\033[36m"
__WHITE = "\033[37m"
__GREY = "\033[90m"

# Defines the colors for the specific verbose types
__LOG = __GREY
__WARNING = __YELLOW
__ERROR = __RED
__SUCCESS = __GREEN
__INFO = __CYAN
__DEBUG = __PURPLE

LOG_VERBOSITY: int = 0
"""Defines the verbosity for all messages coming from the API and requests."""

SUCCESS_VERBOSITY: int = 1
"""Defines the verbosity for all success and errors coming from the API and requests."""

INFO_VERBOSITY: int = 2
"""Defines the verbosity for all information coming from the API and requests."""

WARNING_VERBOSITY: int = 3
"""Defines the verbosity for any warnings or errors coming from the API and requests."""

ERROR_VERBOSITY: int = 4
"""Defines the verbosity for only errors coming from the API and requests."""

# Defines whether the printer should print
__verbose: bool = False
__verbose_level: int = 0

# Defines whether a timestamp should be shown
__display_time: bool = False

# Defines any callbacks
__callbacks: list = []


def set_verbosity(level: int) -> None:
    """
    Updates the verbosity level of the printing to a particular verbosity
    level. The levels are either LOG_VERBOSITY, WARNING_VERBOSITY,
    SUCCESS_VERBOSITY or ERROR_VERBOSITY.

    :param level:   The verbosity level to initialise the printer to
    :type level:    str
    """

    global __verbose, __verbose_level

    if level <= ERROR_VERBOSITY:
        __verbose_level = level
        __verbose = True
    else:
        __verbose_level = 0
        __verbose = False


def clear() -> None:
    """
    Clears the console output, if possible. This is useful for
    clearing the screen before printing new data.
    """

    # Clear the terminal screen
    os.system("cls" if os.name == "nt" else "clear")


def output(data: str, color: str = "") -> None:
    """
    Outputs text to the console assuming that the verbose level
    is enabled for that type.

    :param data:    The raw data to print to the screen
    :type data:     str
    :param color:   The unique character code format for the color or type of text
    :type color:    str
    """

    if __verbose:
        if __display_time:
            data = "[%s] %s" % (__get_time_str(), data)
        print(color + data + __RESET)


def log(data: str) -> None:
    """
    Prints general log text to the console, providing some information
    about the API calls or requests. This requires a LOG_VERBOSITY level.

    :param data:    The raw data to print to the screen
    :type data:     str
    """

    __call_callbacks("log", data)
    if __verbose_level <= LOG_VERBOSITY:
        output(data, __LOG)


def success(data: str) -> None:
    """
    Prints success messages coming from the simulation or the API calls
    to the screen. This requires at least a SUCCESS_VERBOSITY level.

    :param data:    The raw data to print to the screen
    :type data:     str
    """

    __call_callbacks("success", data)
    if __verbose_level <= SUCCESS_VERBOSITY:
        output(data, __SUCCESS)


def info(data: str) -> None:
    """
    Prints information messages to the console, providing information
    about potential results. This requires at least a INFO_VERBOSITY
    level.

    :param data:    The raw data to print to the screen
    :type data:     str
    """

    __call_callbacks("info", data)
    if __verbose_level <= INFO_VERBOSITY:
        output(data, __INFO)


def warning(data: str) -> None:
    """
    Prints warning messages to the console, providing information
    about potential errors. This requires at least a WARNING_VERBOSITY
    level.

    :param data:    The raw data to print to the screen
    :type data:     str
    """

    __call_callbacks("warning", data)
    if __verbose_level <= WARNING_VERBOSITY:
        output(data, __WARNING)


def error(data: str) -> None:
    """
    Prints error messages to the console, providing information about errors
    that occurred. This requires any valid verbosity level.

    :param data:    The raw data to print to the screen
    :type data:     str
    """

    __call_callbacks("error", data)
    if __verbose_level <= ERROR_VERBOSITY:
        output(data, __ERROR)


def debug(data: str) -> None:
    """
    Prints debug messages for testing certain features to the console. This
    will always print provided the printer is enabled, regardless of the
    verbosity level.

    :param data:    The raw data to print to the screen
    :type data:     str
    """

    __call_callbacks("debug", data)
    output(data, __DEBUG)


def display_time(enable: bool) -> None:
    """
    Defines whether time should be enabled on the printing output
    messages or not.

    :param enable:  A flag for enabling the timestamp on the print lines
    :param type:    bool
    """

    global __display_time
    __display_time = enable


def add_callback(callback: callable) -> None:
    """
    Adds a callback function that will be called whenever a message is printed.
    This can be used to log messages or perform other actions when printing.

    :param callback: The callback function to add
    :type callback: callable
    """

    if callable(callback):
        __callbacks.append(callback)
    else:
        raise TypeError("Callback must be a callable function.")


def remove_callback(callback: callable) -> None:
    """
    Removes a callback function from the list of callbacks.

    :param callback: The callback function to remove
    :type callback: callable
    """

    if callback in __callbacks:
        __callbacks.remove(callback)
    else:
        raise ValueError("Callback not found in the list of callbacks.")


def __call_callbacks(type: str, data: str) -> None:
    """
    Calls all registered callback functions with the given data.

    :param type: The type of the message (e.g., "log", "error", etc.)
    :type type: str
    :param data: The data to pass to the callback functions
    :type data: str
    """

    for callback in __callbacks:
        callback(type, data)


def __get_time_str() -> str:
    """
    Returns the current datetime in the suitable datetime
    format that can be printed.

    :returns:   The current time string of the current datatime in the correct format
    :rtype:     str
    """

    return datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f")[:-3]


# Define the base verbosity by default
set_verbosity(ERROR_VERBOSITY)
display_time(True)



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

"""
This modules assists with some helper functions for defining what values
can be turned into JSON and whether GUID IDs are valid. This also includes
some methods for serializing and deserializing JSON data to standard formats.
"""

import re
import numpy as np
from datetime import datetime
from ..utils import ZendirException


def empty_guid() -> str:
    """
    Returns an empty GUID, which is a string of zeros.

    :returns:   An empty GUID
    :rtype:     str
    """

    return "00000000-0000-0000-0000-000000000000"


def is_valid_guid(guid: str) -> bool:
    """
    Determines if a parsed GUID, as a string,
    is valid. This will ensure that it is of
    the correct format.

    :param guid:    The unique GUID of the object, in the correct form
    :type guid:     str

    :returns:       A flag whether the GUID is valid
    :rtype:         bool
    """

    empty: str = empty_guid()
    if guid == None:
        return False
    if guid == empty:
        return False
    if len(guid) != len(empty):
        return False
    if guid[8] != empty[8]:
        return False
    if guid[13] != empty[13]:
        return False
    if guid[18] != empty[18]:
        return False
    if guid[23] != empty[23]:
        return False
    return True


def validate_type(type: str) -> str:
    """
    Validates the type of the object and ensures that it is in the correct format.

    :param type:        The type of the object to validate
    :type type:         str

    :returns:           The validated type with the namespace
    :rtype:             str
    """

    # Check for an invalid or missing
    if type == None or type == "":
        raise ZendirException(
            "Invalid Type: No instance type passed when constructing an instance."
        )

    # Make sure the type is pascal cased
    type = type[0].upper() + type[1:]
    return type


def serialize(value: any) -> any:
    """
    Serializes the value into a JSON serializable format. This will
    convert the value into a list if it is a numpy array or a datetime
    into a string.

    :param value:   The value to serialize
    :type value:    any

    :returns:       The serialized value
    :rtype:         any
    """

    # Check if the value is a numpy array
    if isinstance(value, np.ndarray):
        # Check if the numpy array contains numbers
        if np.issubdtype(value.dtype, np.number):
            return value.tolist()  # Convert to list or list of lists
        return value

    # Check if the value is a datetime
    if isinstance(value, datetime):
        return value.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

    # Check if the value is a simulation instance
    if hasattr(value, "get_id") and hasattr(value, "get_type"):
        return value.get_id()

    # Return the value as is for other types
    return value


def deserialize(value: any) -> any:
    """
    Deserializes the value from a JSON serializable format. This will
    convert the value into a numpy array if it is a list of numbers or
    a datetime string into a datetime object.

    :param value:   The value to deserialize
    :type value:    any

    :returns:       The deserialized value
    :rtype:         any
    """

    # Check if the value is a list
    if isinstance(value, list):

        # Attempt to convert list of numbers to a numpy array
        try:
            array = np.array(value)
            if np.issubdtype(array.dtype, np.number):
                return array
        except:
            pass

        # If conversion fails or is not a list of numbers, return the list as is
        return value

    # Check if the value is a datetime string
    if isinstance(value, str):

        # Regular expression to match the expected datetime format
        datetime_regex = r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{7}Z$"

        # Check if the value matches the expected format
        if re.match(datetime_regex, value):
            # Attempt to parse the datetime
            try:
                # Attempt to parse the datetime
                return datetime.strptime(value[:-2], "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                pass  # If parsing fails, return the original string

    # Check if the value is a dictionary and then deserialize all values
    if isinstance(value, dict):
        for key, val in value.items():
            value[key] = deserialize(val)
        return value

    # Return the value as is for other types
    return value



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

"""
This modules assists with some helper functions for running a certain number of
simulations in parallel, using the asyncio library. This is useful for quickly
running simulations.
"""

import asyncio
from ..connection import Client
from ..simulation import Simulation


def run_simulation(
    client: Client,
    main: callable,
    *args,
    dispose: bool = True,
    **kwargs,
) -> None:
    """
    Run a simulation with the provided client and main function. This is a
    synchronous function that creates a simulation handle and runs the main
    function with the simulation. The main function must have the first
    parameter as the simulation handle, and can take any number of additional
    parameters and keyword arguments.

    :param client: The client to use for the simulation.
    :type client:  Client
    :param main:   The main function to run with the simulation.
    :type main:    callable
    :param args:   Additional positional arguments to pass to the main function.
    :type args:   tuple
    :param dispose: Whether to dispose of the simulation handle after running
    :type dispose: bool
    :param kwargs: Additional keyword arguments to pass to the main function.
    :type kwargs: dict

    :return:       None
    :rtype:        None
    """

    # Define an asynchronous function to run the main function with the simulation
    async def __runner():

        # Create the simulation handle
        simulation: Simulation = await Simulation.create(client)

        # Run the main function with the simulation and additional arguments
        await main(simulation, *args, **kwargs)

        # If dispose is True, dispose of the simulation handle
        if dispose and simulation.is_valid():
            await simulation.dispose()

    # Run the asynchronous function to run the main function
    asyncio.run(__runner())


def run_simulations(
    client: Client, number: int, main: callable, *args, dispose: bool = True, **kwargs
) -> None:
    """
    Run a number of simulations in parallel with the provided client and main
    function. This is a synchronous function that creates a simulation handle
    for each simulation and runs the main function with the simulation. The
    main function must have the first parameter as the simulation handle, the
    second parameter being the index and can take any number of additional
    parameters and keyword arguments.

    :param client: The client to use for the simulations.
    :type client:  Client
    :param number: The number of simulations to run.
    :type number:  int
    :param main:   The main function to run with each simulation.
    :type main:    callable
    :param args:   Additional positional arguments to pass to the main function.
    :type args:    tuple
    :param dispose: Whether to dispose of the simulation handles after running
    :type dispose: bool
    :param kwargs: Additional keyword arguments to pass to the main function.
    :type kwargs: dict

    :return:       None
    :rtype:        None
    """

    # Define an asynchronous function to create a simulation and run the main function
    async def create_simulation(client: Client) -> Simulation:
        return await Simulation.create(client)

    # Define an asynchronous function to run the main function with the simulation
    async def run_and_dispose(sim: Simulation, i, *args, **kwargs):
        try:
            await main(sim, i, *args, **kwargs)
        finally:
            if dispose and sim.is_valid():
                await sim.dispose()

    # Define an asynchronous function to run all simulations concurrently
    async def run_all():
        # Create all simulations concurrently
        simulations = await asyncio.gather(
            *(create_simulation(client) for _ in range(number))
        )
        # Run the main function for each simulation concurrently, then dispose
        tasks = [
            asyncio.create_task(run_and_dispose(sim, i, *args, **kwargs))
            for i, sim in enumerate(simulations)
        ]
        await asyncio.gather(*tasks)

    # Run the asynchronous function to run all simulations
    asyncio.run(run_all())



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

"""
This module contains public methods that aid with mathematical
calculations and astrophysical dimensions.
"""

import numpy as np


def acos_quadrant_check(adjacent: float, hypotenuse: float, test: float) -> float:
    """
    Calculate an angle using the arccosine function and perform a
    quadrant check. The function calculates the angle using the arccosine
    of the ratio of the adjacent side to the hypotenuse. It adjusts the
    angle based on the quadrant determined by the 'test' parameter.

    :param adjacent:    The length of the adjacent side of a right triangle.
    :type adjacent:     float
    :param hypotenuse:  The length of the hypotenuse of the triangle.
    :type hypotenuse:   float
    :param test:        A value to determine the quadrant of the angle.
    :type test:         float

    :returns:           The calculated angle in radians.
    :rtype:             float
    """
    # Calculate the ratio of the adjacent side to the hypotenuse
    rat = adjacent / hypotenuse

    # Adjust the ratio if the adjacent side is slightly greater than the hypotenuse
    if adjacent > hypotenuse:
        if np.fabs((adjacent - hypotenuse) / hypotenuse) < 1e-10:
            rat = 1.0
        else:
            raise ValueError(
                "The adjacent: {}, is greater than the hypotenuse: {}".format(
                    adjacent, hypotenuse
                )
            )
    # Calculate the angle using arccos; adjust based on the 'test' parameter
    return 2 * np.pi - np.arccos(rat) if test < 0 else np.arccos(rat)


def normalize_angle(
    angle: float | np.ndarray, angle_max: float = 2 * np.pi
) -> float | np.ndarray:
    """
    Normalize an angle or array of angles between 0 <= theta < angle_max, where angle_max is an arbitrary angle.

    :param angle: The angle (float) or array of angles (numpy.ndarray) to be normalized.
    :type angle: float or numpy.ndarray
    :param angle_max: The maximum value for the angle range (default: 2*pi for radians).
    :type angle_max: float

    :returns: The normalized angle (float) or array of angles (numpy.ndarray).
    :rtype: float or numpy.ndarray

    :example:
    >>> normalize_angle(-3.0, angle_max=360)
    357.0
    >>> normalize_angle(np.array([-3.0, 0.0, 6.5, 12.0, -9.0]), angle_max=360)
    array([357. ,   0. ,   6.5,  12. , 351. ])
    """
    angle = np.fmod(angle, angle_max)
    # Add the angle max if the angle is negative. Handle for both scalar and array angles
    if np.isscalar(angle):
        if angle < 0:
            angle += angle_max
    else:
        angle[angle < 0] += angle_max
    return angle


def shortest_angular_difference(angle1: float, angle2: float) -> float:
    """
    Calculate the shortest angular difference between two angles. This function
    calculates the shortest angular difference between two angles using the
    arccosine function and the quadrant check function.

    :param angle1:  The first angle
    :type angle1:   float
    :param angle2:  The second angle
    :type angle2:   float

    :returns:       The shortest angular difference between the two angles
    :rtype:         float
    """
    # Calculate the difference between the two angles
    d_angle: float = angle1 - angle2
    # Get the alternative angle
    alt_angle: float = -np.sign(d_angle) * (2 * np.pi - np.fabs(d_angle))
    # Return the shortest angle
    return alt_angle if np.fabs(alt_angle) < np.fabs(d_angle) else d_angle


def normalize_array(array: np.ndarray) -> np.ndarray:
    """
    Normalize a numpy array. This function normalizes a numpy array by
    dividing each element by the magnitude of the array.

    :param array:   The array to be normalized
    :type array:    np.ndarray

    :returns:       The normalized array
    :rtype:         np.ndarray
    """
    # Return the normalized array
    return array / np.linalg.norm(array)


def parallel_vector(v1: np.ndarray, v_ref: np.ndarray) -> np.ndarray:
    """
    Calculate the component of v1 that is parallel with v_ref.

    :param v1:  The first vector
    :type v1:   np.ndarray
    :param v_ref:  The second vector
    :type v_ref:   np.ndarray

    :returns:   The parallel vector of the two vectors
    :rtype:     np.ndarray
    """
    # find the parrallel component
    reference_normalised = normalize_array(v_ref)
    return reference_normalised * np.dot(v1, reference_normalised)


def perpendicular_vector(v1: np.ndarray, v_ref: np.ndarray) -> np.ndarray:
    """
    Calculate the perpendicular vector of two vectors. This function calculates
    the perpendicular vector of two vectors using the dot product and the magnitude
    of each vector. The perpendicular vector is calculated using the dot product
    and the magnitude of the first vector.

    :param v1:  The first vector
    :type v1:   np.ndarray
    :param v_ref:  The second vector
    :type v_ref:   np.ndarray

    :returns:   The perpendicular vector of the two vectors
    :rtype:     np.ndarray
    """
    # find the perpendicular component
    return v1 - parallel_vector(v1, v_ref)


def angle_between_vectors(v1: np.ndarray, v2: np.ndarray) -> float:
    """
    Calculate the angle between two vectors. This function calculates the
    angle between two vectors using the dot product and the magnitude of
    each vector. The angle is calculated using the arccosine function.

    :param v1:  The first vector
    :type v1:   np.ndarray
    :param v2:  The second vector
    :type v2:   np.ndarray

    :returns:   The angle between the two vectors in radians.
    :rtype:     float
    """

    # Calculate the dot product of the two vectors
    dot: float = np.dot(v1, v2)

    # Calculate the magnitude of each vector
    denom: float = np.linalg.norm(v1) * np.linalg.norm(v2)
    if denom == 0:
        raise ZeroDivisionError
    # Calculate the angle between the two vectors
    angle: float = acos_quadrant_check(dot, denom, dot)

    return angle


def unit_vector_derivative(u: np.ndarray, udot: np.ndarray) -> np.ndarray:
    """
    Calculate the time derivative of a unit vector.

    :param u: The original vector.
    :type u: numpy.ndarray
    :param udot: The time derivative of the original vector.
    :type udot: numpy.ndarray
    :return: The time derivative of the unit vector derived from the original vector.
    :rtype: numpy.ndarray

    Computes the time derivative of a unit vector given a vector and its time derivative.
    If the magnitude of the original vector is zero, returns a zero vector to avoid division by zero.

    Example:
    --------
    >>> u = np.array([1, 0, 0])
    >>> udot = np.array([0, 1, 0])
    >>> unit_vector_derivative(u, udot)
    array([0., 1., 0.])
    """
    umag: float = np.linalg.norm(u)
    if umag == 0:
        uhatd: np.ndarray = np.zeros(3)
    else:
        uhat: np.ndarray = u / umag
        uhatd: np.ndarray = (udot - np.dot(uhat, udot) * uhat) / umag
    return uhatd



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

import paho.mqtt.client as mqtt
from typing import Callable
import uuid as Guid
import json, time
from ..utils import printer


# These are some static constants for the client
HIVEMQ_SERVER: str = "broker.hivemq.com"
"""The default MQTT server that is being used to communicate with the clients."""

SERVER_PORT: int = 1883
"""The default MQTT server port that is being used for the protocol."""

SERVER_TIMEOUT: int = 5
"""The server timeout in seconds to keep the connection alive for."""


class Client:
    """
    This client acts as an MQTT interface with python, allowing for
    callbacks from topic subscriptions and publishing messages over
    topics. This acts as a wrapper for the Paho MQTT client, which
    must be installed to work correctly.
    """

    server: str = ""
    """Defines the name of the server being used to communicate with."""

    client_name: str = ""
    """Defines the unique name of the client."""

    connected: bool = False
    """A flag for if the MQTT client is connected to the server."""

    fail: bool = False
    """A flag for if a failure to communicate with the server occurred."""

    callbacks: dict = {}
    """A mapping of topic to functions to callback as a dictionary."""

    def __init__(self, server: str = HIVEMQ_SERVER) -> None:
        """
        Initialiases the client with a particular server, defaulted
        to the HiveMQT server. This will also settup the delegate
        function calls.

        :param server:  The name of the server to connect to
        :type server:   str
        """

        self.server = server
        self.client_name = str(Guid.uuid4())

        # Create the client
        self.client = mqtt.Client(self.client_name)
        self.client.on_connect = self.__on_connect
        self.client.on_message = self.__on_message
        self.client.on_connect_fail = self.__on_connect_fail

    def add_callback(self, topic: str, func: Callable[[str], None]) -> None:
        """
        Adds a particular function callback to the data when received over
        a topic. The callback function must be in the form of 'def func (data: str)'.
        Multiple functions can be added to each topic and will execute each
        of the functions with the same data.

        :param topic:   The topic name that should have a callback with
        :type topic:    str
        :param func:    The callback function when the topic receives data, in the form of func (data: str)
        :type func:     Callable
        """

        # Ensure the client has already connected
        if not self.connected:
            printer.error(
                "Failed to subscribe to topic. Mqtt client is not yet connected."
            )
            return

        # Check if the callback exists
        if topic in self.callbacks.keys():
            list(self.callbacks[topic]).append(func)

        # Add a new callback array
        else:
            self.client.subscribe(topic)
            self.callbacks[topic] = [func]

        # Add a message
        printer.log("Successfully subscribed to topic '%s'." % topic)

    def connect(self, wait: bool = False) -> None:
        """
        This method begins to connect to the server and makes
        sure that it will run until shutdown. If the wait flag
        is enabled, it will throttle the thread until a valid
        connection has been made.

        :param wait:    A flag whether to wait until the connection is reached and pause the thread
        :type wait:     bool
        """

        # Connect to the server
        self.client.connect(self.server, SERVER_PORT, SERVER_TIMEOUT)

        # Loop until connected
        printer.log(
            "Mqtt client '%s' is attempting to connect to the server..."
            % self.client_name
        )
        self.client.loop_start()

        # Wait for the connection if required
        if wait:
            self.wait_for_connection()

    def wait_for_connection(self, timeout: float = 3.0) -> bool:
        """
        Waits for a valid connection after calling connect. This will wait
        for a certain time until it time's out and returns out of the loop.

        :param timeout: The timeout in seconds for the connection to wait for
        :type timeout:  float

        :returns:       A valid connection flag if a connection has been reached
        :rtype:         bool
        """

        ticks: int = int(float(timeout) / 0.1)
        for _ in range(ticks):
            if self.connected:
                return True
            time.sleep(0.1)
        printer.warning("Failed to connect with a timeout of %f seconds." % timeout)
        return False

    def __on_connect(self, client, userdata, flags, rc):
        """
        This handles the callback from the connection and
        will display a message outlining if it has been connected.
        """

        printer.log("Mqtt client has connected with result code: '%s'." % str(rc))
        self.connected = True

    def __on_connect_fail(self):
        """
        This method is executed if the connection with a server fails
        and is unable to connect.
        """

        printer.error("Mqtt client has failed to connect.")
        self.fail = True

    def __on_message(self, client, userdata, message):
        """
        Handles when a particular message has come through
        the system. This will redirect a message to a callback
        unless otherwise noted.
        """

        # Check if the topic is valid
        if message.topic not in self.callbacks.keys():
            printer.warning("Undefined message on topic '%s'." % message.topic)

        # Execute each of the callbacks
        else:

            # Fetch the data as a string or a byte array
            data = None
            try:
                data = message.payload.decode()
            except:
                data = message.payload

            # Execute each of the functions with the data
            for func in self.callbacks[message.topic]:
                func(data)

    def publish(self, topic: str, data: str) -> None:
        """
        Publishes some data over a particular topic, assuming it is
        string data.

        :param topic:   The topic to publish the data over the MQTT connection
        :type topic:    str
        :param data:    The data to publish over the MQTT topic
        :type data:     str
        """

        self.client.publish(topic, payload=data)
        printer.log("Published a message over the topic '%s'." % topic)

    def publish_json(self, topic: str, data: dict) -> None:
        """
        Publishes some JSON data over a particular topic, assuming it
        is in the form of a dictionary or some other JSON configurable
        object.

        :param topic:   The topic to publish the data over the MQTT connection
        :type topic:    str
        :param data:    The JSON value to publish over the MQTT topic
        :type data:     dict
        """

        data_str: str = json.dumps(data)
        self.publish(topic, data_str)



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from ..utils import printer, ZendirException, helper
from .instance import Instance
from .message import Message
from .context import Context


class System(Instance):
    """
    A system is a singleton object within the simulation. It only exists once
    per simulation and is used to define the global state of the simulation.
    """

    __messages: dict[str:Message] = {}
    """Defines all messages that are attached to the object, by name."""

    def __init__(self, context: Context, id: str, type: str = None) -> None:
        """
        Initialises the system with the context and the ID of the system.

        :param context:         The context of the simulation
        :type context:          Context
        :param id:              The GUID ID of the system
        :type id:               str
        :param type:            The type of the system, if applicable
        :type type:             str
        """

        super().__init__(context, id, type)

        # Initialised the base data
        self.__messages = {}

    def _reset_refresh_cache(self) -> None:
        """
        Overrides the base class method to set the flag for refreshing the cache to true.
        This will ensure that all messages will also require a refresh.
        """

        # Ensure all messages require a refresh too
        for message in self.__messages.values():
            message._reset_refresh_cache()
        super()._reset_refresh_cache()

    def get_instance_with_id(self, id: str, recurse: bool = False) -> Instance:
        """
        Returns the instance that is attached to this behaviour with the specified ID. If the
        instance does not exist, None will be returned. This may be a message or a child instance
        of some kind.

        :param id:  The ID of the instance to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of children to find the instance
        :type recurse:   bool

        :returns:   The instance that is attached to the object with the specified ID
        :rtype:     Instance
        """

        # Start by looking at the instances for the ID
        for msg in self.__messages.values():
            if msg.get_id() == id:
                return msg

        # Return none otherwise
        return None

    async def get_message(self, name: str) -> Message:
        """
        Attempts to get the message with the specified name that is attached to the system. If the
        message does not exist, it will be created and attached to the system. If the message cannot
        be created, an exception will be raised.

        :param name:    The name of the message to fetch
        :type name:     str

        :returns:       The message with the specified name that is attached to the system
        :rtype:         Message
        """

        # Check if the name is within the message structure and return that
        if name in self.__messages.keys():
            return self.__messages[name]

        # Fetch the data
        message_id: str = await self.get(name)
        if not helper.is_valid_guid(message_id):
            raise ZendirException(f"Failed to find message with name '{name}'.")

        # Create the message object with the ID
        message = Message(self._context, message_id)
        self.__messages[name] = message

        # Return the message of that name
        printer.success(f"Successfully created message with name '{name}'.")
        return message

    def get_messages(self) -> list[Message]:
        """
        Returns all of the messages that are attached to the system. This will only include the
        messages that have currently been fetched.

        :returns:   All of the messages that are attached to the system
        :rtype:     list[Message]
        """

        return list(self.__messages.values())



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from . import printer


class ZendirException(Exception):
    """
    The Zendir Exception class defines a custom exception that is able
    to track errors with the Zendir API. Any connection errors or issues
    with the objects will throw an Zendir API if there are configuration
    issues.
    """

    def __init__(self, message: str):
        """
        Defines the constructor for the exception and is able to pass in
        some parameters in regards to the exception. This will also print
        the exception message

        :param message:     The information error message that will be thrown.
        :type message:      str
        """

        printer.error(message)
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        """
        This is the automated cast to a string from the message, which is
        able to return the exception information.

        :returns:   The string formatted text of the message
        :rtype:     str
        """

        return "[ZENDIR ERROR] %s" % self.message



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

"""
This file contains public constants related to general mathematics,
as well as astrophysical constants. These include planetary data
and useful values that can be used for conversions.
"""

M_PI: float = 3.141592653589793
"""[-] The value of PI"""

M_PI_2: float = 1.5707963267948966
"""[-] The value of PI/2"""

M_E: float = 2.718281828459045
"""[-] The value of the exponential constant E"""

D2R: float = M_PI / 180.0
"""[-] Degrees to Radians constant"""

R2D: float = 180.0 / M_PI
"""[-] Radians to Degrees constant"""

RPM: float = 0.10471975511965977
"""[-] Conversion from rad/s to revolutions per minute"""

SEC2DAY: float = 1.0 / (60.0 * 60.0 * 24.0)
"""[d] The number of days in a second"""


G: float = 6.67259e-11
"""[m^3/s^2/kg] Universial Gravitational Constant"""

AU: float = 149597870693.0
"""[m] Astronomical unit"""

LIGHT_SPEED: float = 299792458
"""[m/s] The absolute speed of light in a vacuum in meters per second"""

CELCIUS_TO_KELVIN: float = 273.15
"""[K] The value of temperature in Kelvin at 0 degrees Celcius"""


SUN_REQ: float = 695000000
"""[m] The equatorial radius of the Sun"""

SUN_RP: float = 6.95508e8
"""[m] Polar radius of the Sun"""

SUN_MASS: float = 1.989e30
"""[kg] The total mass of the Sun"""

SUN_J2: float = 0.0
"""[-] The Sun J2 value (implicitely zero)"""

SUN_MU: float = 132712440023.310 * 1e9
"""[m^3/s^2] The Sun Gravitational constant"""

SUN_DENSITY: float = 1410
"""[kg/m^3] The density of Sun"""

SUN_ALBEDO_AVG: float = 0.0
"""[-] A reflection albedo constant from the Sun's surface"""


MERCURY_REQ: float = 2439700
"""[m] Equatorial radius of Mercury"""

MERCURY_RP: float = 2.4397e6
"""[m] Polar radius of Mercury"""

MERCURY_MASS: float = 0.330e24
"""[kg] The total mass of Mercury"""

MERCURY_J2: float = 60.0e-6
"""[-] Mercury J2 value"""

MERCURY_ORBIT_SMA: float = 0.38709893 * AU
"""[m] Mercury Semi-Major Axis of it's orbit"""

MERCURY_ORBIT_INC: float = 7.00487 * D2R
"""[rad] The axis inclination of Mercury's orbit"""

MERCURY_ORBIT_E: float = 0.20563069
"""[-] The eccentricity for Mercury's orbit"""

MERCURY_OMEGA: float = 2 * M_PI / 1407.6 / 3600.0
"""[r/s] Rotation rate of the planet"""

MERCURY_MU: float = 22032.080 * 1e9
"""[m^3/s^2] The gravitational constant for Mercury"""

MERCURY_DENSITY: float = 5427
"""[kg/m^3] The density of Mercury"""

MERCURY_ALBEDO_AVG: float = 0.119
"""[-] A reflection albedo constant from Mercury's surface"""


VENUS_REQ: float = 6051800
"""[m] Equatorial radius of Venus"""

VENUS_RP: float = 6.0518e6
"""[m] Polar radius of Venus"""

VENUS_MASS: float = 4.87e24
"""[kg] The total mass of Venus"""

VENUS_J2: float = 4.458e-6
"""[-] Venus J2 value"""

VENUS_ORBIT_SMA: float = 0.72333199 * AU
"""[m] Venus Semi-Major Axis of its orbit"""

VENUS_ORBIT_INC: float = 3.39471 * D2R
"""[rad] The axis inclination of Venus's orbit"""

VENUS_ORBIT_E: float = 0.00677323
"""[-] The eccentricity for Venus's orbit"""

VENUS_OMEGA: float = 2 * M_PI / -5832.6 / 3600.0
"""[r/s] Rotation rate of the planet"""

VENUS_MU: float = 324858.599 * 1e9
"""[k=m^3/s^2] The gravitational constant for Venus"""

VENUS_DENSITY: float = 5243
"""[kg/m^3] The density of Venus"""

VENUS_ALBEDO_AVG: float = 0.75
"""[-] A reflection albedo constant from Venus's surface"""


EARTH_REQ: float = 6378136.6
"""[m] Equatorial radius of Earth"""

EARTH_FLATTENING: float = 1.0 / 298.257223563
"""[-] The flattening of the Earth"""

EARTH_RP: float = 6356751.9
"""[m] Polar radius of Earth"""

EARTH_MASS: float = 5.97e24
"""[kg] The total mass of Earth"""

EARTH_J2: float = 1082.616e-6
"""[-] Earth J2 value"""

EARTH_J3: float = -2.53881e-6
"""[-] Earth J3 value"""

EARTH_J4: float = -1.65597e-6
"""[-] Earth J4 value"""

EARTH_J5: float = -0.15e-6
"""[-] Earth J5 value"""

EARTH_J6: float = 0.57e-6
"""[-] Earth J6 value"""

EARTH_ORBIT_SMA: float = 1.00000011 * AU
"""[m] Earth Semi-Major Axis of its orbit"""

EARTH_ORBIT_INC: float = 0.00005 * D2R
"""[rad] The axis inclination of Earth's orbit"""

EARTH_ORBIT_E: float = 0.01671022
"""[-] The eccentricity for Earth's orbit"""

EARTH_OMEGA: float = 2 * M_PI / 23.9345 / 3600.0
"""[rad/s] Rotation rate of the planet"""

EARTH_MU: float = 3.986004414e14
"""[m^3/s^2] The gravitational constant for Earth"""

EARTH_DENSITY: float = 5514
"""[kg/m^3] The density of Earth"""

EARTH_ALBEDO_AVG: float = 0.29
"""[-] A reflection albedo constant from Earth's surface"""

EARTH_SOLAR_FLUX: float = 1372.5398
"""[W/m^2] Median solar flux at Earth"""

EARTH_GRAV: float = 9.80665
"""[m/s] The gravitational acceleration at the surface of the Earth"""

MOON_REQ: float = 1737400
"""[m] Equatorial radius of the Moon"""

MOON_RP: float = 1.7374e6
"""[m] Polar radius of the Moon"""

MOON_MASS: float = 0.073e24
"""[kg] The total mass of the Moon"""

MOON_J2: float = 202.7e-6
"""[-] J2 constant for the Moon"""

MOON_ORBIT_SMA: float = 0.3844e9
"""[m] Semi-Major Axis of the Moon's orbit"""

MOON_ORBIT_E: float = 0.0549
"""[-] Eccentricity of the Moon's orbit"""

MOON_OMEGA: float = 2 * M_PI / 655.728 / 3600.0
"""[rad/s] The Moon's rotation rate"""

MOON_MU: float = 4.9048695e12
"""[m^3/s^2] The gravitational constant for the Moon"""

MOON_DENSITY: float = 3340
"""[kg/m^3] The density of the Moon"""

MOON_ALBEDO_AVG: float = 0.123
"""[-] A reflection albedo constant from the Moon's surface"""


MARS_REQ: float = 3396190
"""[m] Equatorial radius of Mars"""

MARS_RP: float = 3376200
"""[m] Polar radius of Mars"""

MARS_MASS: float = 0.642e24
"""[kg] The total mass of Mars"""

MARS_J2: float = 1960.45e-6
"""[-] Mars J2 value"""

MARS_ORBIT_SMA: float = 1.52366231 * AU
"""[m] Mars Semi-Major Axis of its orbit"""

MARS_ORBIT_INC: float = 1.85061 * D2R
"""[rad] The axis inclination of Mars's orbit"""

MARS_ORBIT_E: float = 0.09341233
"""[-] The eccentricity for Mars's orbit"""

MARS_OMEGA: float = 2 * M_PI / 24.6229 / 3600.0
"""[r/s] Rotation rate of the planet"""

MARS_MU: float = 42828.314 * 1e9
"""[m^3/s^2] The gravitational constant for Mars"""

MARS_DENSITY: float = 3933
"""[kg/m^3] The density of Mars"""

MARS_ALBEDO_AVG: float = 0.16
"""[-] A reflection albedo constant from Mars's surface"""

PHOBOS_REQ: float = 11200
"""[m] The radius of Phobos"""

DEIMOS_REQ: float = 6100
"""[m] The radius of Deimos"""


JUPITER_REQ: float = 71492000
"""[m] Equatorial radius of Jupiter"""

JUPITER_RP: float = 7.1492e7
"""[m] Polar radius of Jupiter"""

JUPITER_MASS: float = 1898e24
"""[kg] The total mass of Jupiter"""

JUPITER_J2: float = 14736.0e-6
"""[-] Jupiter J2 value"""

JUPITER_ORBIT_SMA: float = 5.20336301 * AU
"""[m] Jupiter Semi-Major Axis of its orbit"""

JUPITER_ORBIT_INC: float = 1.30530 * D2R
"""[rad] The axis inclination of Jupiter's orbit"""

JUPITER_ORBIT_E: float = 0.04839266
"""[-] The eccentricity for Jupiter's orbit"""

JUPITER_OMEGA: float = 2 * M_PI / 9.9250 / 3600.0
"""[r/s] Rotation rate of the planet"""

JUPITER_MU: float = 126712767.881 * 1e9
"""[m^3/s^2] The gravitational constant for Jupiter"""

JUPITER_DENSITY: float = 1326
"""[kg/m^3] The density of Jupiter"""

JUPITER_ALBEDO_AVG: float = 0.34
"""[-] A reflection albedo constant from Jupiter's surface"""


SATURN_REQ: float = 60268000
"""[m] Equatorial radius of Saturn"""

SATURN_RP: float = 6.0268e7
"""[m] Polar radius of Saturn"""

SATURN_MASS: float = 568e24
"""[kg] The total mass of Saturn"""

SATURN_J2: float = 16298.0e-6
"""[-] Saturn J2 value"""

SATURN_ORBIT_SMA: float = 9.53707032 * AU
"""[m] Saturn Semi-Major Axis of its orbit"""

SATURN_ORBIT_INC: float = 2.48446 * D2R
"""[rad] The axis inclination of Saturn's orbit"""

SATURN_ORBIT_E: float = 0.05415060
"""[-] The eccentricity for Saturn's orbit"""

SATURN_OMEGA: float = 2 * M_PI / 10.656 / 3600.0
"""[r/s] Rotation rate of the planet"""

SATURN_MU: float = 37940626.068 * 1e9
"""[m^3/s^2] The gravitational constant for Saturn"""

SATURN_DENSITY: float = 687
"""[kg/m^3] The density of Saturn"""

SATURN_ALBEDO_AVG: float = 0.34
"""[-] A reflection albedo constant from Saturn's surface"""


URANUS_REQ: float = 25559000
"""[m] Equatorial radius of Uranus"""

URANUS_RP: float = 2.5559e7
"""[m] Polar radius of Uranus"""

URANUS_MASS: float = 86.8e24
"""[kg] The total mass of Uranus"""

URANUS_J2: float = 3343.43e-6
"""[-] Uranus J2 value"""

URANUS_ORBIT_SMA: float = 19.19126393 * AU
"""[m] Uranus Semi-Major Axis of its orbit"""

URANUS_ORBIT_INC: float = 0.76986 * D2R
"""[rad] The axis inclination of Uranus's orbit"""

URANUS_ORBIT_E: float = 0.04716771
"""[-] The eccentricity for Uranus's orbit"""

URANUS_OMEGA: float = 2 * M_PI / -17.24 / 3600.0
"""[r/s] Rotation rate of the planet"""

URANUS_MU: float = 5794559.128 * 1e9
"""[m^3/s^2] The gravitational constant for Uranus"""

URANUS_DENSITY: float = 1271
"""[kg/m^3] The density of Uranus"""

URANUS_ALBEDO_AVG: float = 0.29
"""[-] A reflection albedo constant from Uranus's surface"""


NEPTUNE_REQ: float = 24746000
"""[m] Equatorial radius of Neptune"""

NEPTUNE_RP: float = 2.4746e7
"""[m] Polar radius of Neptune"""

NEPTUNE_MASS: float = 102e24
"""[kg] The total mass of Neptune"""

NEPTUNE_J2: float = 3411.0e-6
"""[-] Neptune J2 value"""

NEPTUNE_ORBIT_SMA: float = 30.06896348 * AU
"""[m] Neptune Semi-Major Axis of its orbit"""

NEPTUNE_ORBIT_INC: float = 1.76917 * D2R
"""[rad] The axis inclination of Neptune's orbit"""

NEPTUNE_ORBIT_E: float = 0.00858587
"""[-] The eccentricity for Neptune's orbit"""

NEPTUNE_OMEGA: float = 2 * M_PI / 16.11 / 3600.0
"""[r/s] Rotation rate of the planet"""

NEPTUNE_MU: float = 6836534.065 * 1e9
"""[m^3/s^2] The gravitational constant for Neptune"""

NEPTUNE_DENSITY: float = 1638
"""[kg/m^3] The density of Neptune"""

NEPTUNE_ALBEDO_AVG: float = 0.31
"""[-] A reflection albedo constant from Neptune's surface"""


PLUTO_REQ: float = 1137000
"""[m] Equatorial radius of Pluto"""

PLUTO_RP: float = 1.37e6
"""[m] Polar radius of Pluto"""

PLUTO_MASS: float = 0.0146e24
"""[kg] The total mass of Pluto"""

PLUTO_J2: float = 0.0
"""[-] Pluto J2 value"""

PLUTO_ORBIT_SMA: float = 39.48168677 * AU
"""[m] Pluto Semi-Major Axis of its orbit"""

PLUTO_ORBIT_INC: float = 17.14175 * D2R
"""[rad] The axis inclination of Pluto's orbit"""

PLUTO_ORBIT_ECC: float = 0.24880766
"""[-] The eccentricity for Pluto's orbit"""

PLUTO_OMEGA: float = 2 * M_PI / -153.2928 / 3600.0
"""[r/s] Rotation rate of the planet"""

PLUTO_MU: float = 983.055 * 1e9
"""[m^3/s^2] The gravitational constant for Pluto"""

PLUTO_DENSITY: float = 2095
"""[kg/m^3] The density of Pluto"""

PLUTO_ALBEDO_AVG: float = 0.0
"""[-] A reflection albedo constant from Pluto's surface"""



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

import json


def create_command(command: str, parameters: dict, time: float = 0.0) -> dict:
    """
    Creates a command with a list of arguments as the parameters. This will create the JSON
    object with the list of parameters required.

    :param command: The command to execute.
    :type command: str
    :param parameters: The parameter dictionary containing keys and values for the command.
    :type parameters: dict
    :param time: The time [s] that the command must be executed at, default is 0.0.
    :type time: float, optional
    :return: The JSON command that can be used.
    :rtype: dict
    """
    # Initialize an empty command dictionary
    cmd = {}

    # Set the 'command' field of the JSON object
    cmd["command"] = command

    # Initialize an empty dictionary for parameters
    param_obj = {}

    # Iterate through the parameters and add them to the dictionary
    for key, value in parameters.items():
        param_obj[key] = value

    # Set the 'parameters' field of the JSON object
    cmd["parameters"] = param_obj

    # Add the timestamp to the JSON object
    cmd["_t"] = time

    # Return the complete command object (JSON)
    return cmd


def create_guidance_command(
    navigation: str = "Simple",
    pointing: str = "Inertial",
    controller: str = "MRP",
    mapping: str = "ReactionWheels",
    time: float = 0.0,
) -> dict:
    """
    Creates a guidance command for the spacecraft operation computer to execute
    once loaded into the system at a particular time.

    :param navigation: The navigation mode to set (default is "Simple").
    :type navigation: str, optional
    :param pointing: The pointing mode to set (default is "Inertial").
    :type pointing: str, optional
    :param controller: The controller mode to set (default is "MRP").
    :type controller: str, optional
    :param mapping: The mapping mode to set (default is "ReactionWheels").
    :type mapping: str, optional
    :param time: The time [s] at which the command is executed, default is 0.0.
    :type time: float, optional
    :return: The appropriate command JSON to be executed.
    :rtype: dict
    """
    # Create a dictionary of arguments with navigation, pointing, controller, and mapping types
    args = {
        "navigation": navigation,
        "pointing": pointing,
        "controller": controller,
        "mapping": mapping,
    }

    # Return the generated command by calling create_command with the arguments
    return create_command("guidance", args, time)


def create_guidance_command_string(
    navigation: str, pointing: str, controller: str, mapping: str, time: float
) -> str:
    """
    Creates a guidance command for the spacecraft operation computer and returns it as a JSON string.

    :param navigation: The navigation mode to set.
    :type navigation: str
    :param pointing: The pointing mode to set.
    :type pointing: str
    :param controller: The controller mode to set.
    :type controller: str
    :param mapping: The mapping mode to set.
    :type mapping: str
    :param time: The time [s] at which the command is executed.
    :type time: float
    :return: The JSON command as a string.
    :rtype: str
    """
    # Generate the command as a dictionary
    command = create_guidance_command(navigation, pointing, controller, mapping, time)

    # Convert the command dictionary to a JSON string and return it
    return json.dumps(command)



"""
                    [ ZENDIR ]
This code is developed by Zendir to aid with communication
to the public API. All code is under the the license provided
with the 'zendir' module. Copyright Zendir, 2025.
"""

from typing import Optional, Dict, Union
from zendir.utils import printer, ZendirException, helper
import asyncio, aiohttp, json, atexit
import requests, time
from typing import Optional, Union, Dict, Any, List
from zendir import __version__ as zendir_version


class Client:
    """
    A simple client to handle HTTP requests to an API.
    Each simulation ID processes requests sequentially using its own session.
    """

    def __init__(
        self,
        url: str = "https://api.zendir.io/v2.0",
        token: Optional[str] = None,
        timeout: Optional[float] = 30,
    ):
        """
        Initialize the client with a base URL and optional token.

        :param url: The base URL for the API.
        :param token: Authentication token for requests.
        """
        self.base_url = url.rstrip("/") + "/"
        self.url = ""
        self.session = ""
        self.token = token
        self.timeout = timeout

        # Fetch the session token if provided
        if self.token != None:

            # Fetch the session information
            self.session_info: List[dict] = self.__get_session_info()

            # If there is an active session, set the URL
            if len(self.session_info) > 0:

                # Find the first session that has a 'version' that matches the version
                version: str = self.get_version()
                valid_sessions: List[str] = [
                    session["guid"]
                    for session in self.session_info
                    if ("version" not in session or session["version"] == version)
                    and helper.is_valid_guid(session["guid"])
                ]

                # If there are no valid sessions, use the first session
                if len(valid_sessions) == 0:

                    # Try to create a new session with the current version
                    try:
                        session_id: str = self.create_session()

                    # If the session, creation fails
                    except Exception as e:

                        # Ask the user if they would like to delete the old sessions
                        input: str = input(
                            "An session with a different version exists. Would you like to delete it? (y/n): "
                        )

                        # If the user wants to delete the old sessions, delete all sessions and then create a new session
                        if input.lower() == "y":
                            for session in self.session_info:
                                self.delete_session(session["guid"])
                            session_id: str = self.create_session()

                        # Otherwise, just use the first session
                        else:
                            session_id = self.session_info[0]["guid"]
                            printer.warning(
                                "Using the first session with a different version. This may cause compatability issues."
                            )

                # If more than one valid session, use the first valid session that is in a 'RUNNING' state
                elif len(valid_sessions) > 1:
                    session_id: str = next(
                        (
                            session["guid"]
                            for session in self.session_info
                            if session["guid"] in valid_sessions
                            and session["status"] == "RUNNING"
                        ),
                        valid_sessions[0],
                    )
                    if not session_id:
                        session_id = valid_sessions[0]

                # Otherwise, use the first valid session
                else:
                    session_id: str = valid_sessions[0]

                # Wait for the session to become active
                self.__wait_for_session(session_id)

                # Session is good to go, set the URL
                self.url = f"{self.base_url}{session_id}/"
                printer.success(
                    f"Successfully connected to the API session '{session_id}'."
                )

            # If there are no active sessions, create a new session
            else:

                # Create a new session and wait for it to become active
                session: str = self.create_session()
                self.__wait_for_session(session)

                # Session is good to go, set the URL
                self.url = f"{self.base_url}{session}/"
                printer.success(
                    f"Successfully created and connected to a new API session '{session}'."
                )

        # If there is no token, set the URL to the base URL
        else:
            self.url = self.base_url
            if "127.0.0.1" not in self.url and "localhost" not in self.url:
                printer.warning(
                    "No token provided. Using the base URL for API requests. "
                    "This may not work as expected without authentication."
                )

    async def _request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Union[str, list, dict]] = None,
    ):
        """
        Make an HTTP request for a given simulation ID.

        :param method: HTTP method (GET, POST, etc.).
        :param endpoint: API endpoint.
        :param data: Data to send with the request.
        :return: Response data as a dictionary or string.
        """

        # Creates the data and headers for the request
        try:
            body = None
            headers = {}
            if data:
                if isinstance(data, str):
                    body = data.encode("utf-8")
                    headers["Content-Type"] = "text/plain"
                    headers["Content-Length"] = str(len(body))
                elif isinstance(data, (list, dict)):
                    body = json.dumps(data).encode()
                    headers["Content-Type"] = "application/json"
                    headers["Content-Length"] = str(len(body))

        # Catch any exceptions that occur during data preparation
        except Exception as e:
            raise ZendirException(f"Failed to prepare request data: {e}")

        # Add the token to the headers if it exists
        if self.token:
            headers["X-Api-Key"] = self.token

        # Define the URL for the request
        url = f"{self.url}{endpoint.lstrip('/')}"

        # Print the request details for debugging
        printer.log(f"Requesting {method} {url} with data: {data}")

        # Open a new session
        async with aiohttp.ClientSession() as session:

            # Create a request and read the response
            async with session.request(
                method, url, data=body, headers=headers, timeout=self.timeout
            ) as response:

                # Check the status of the response
                if response.status != 200:
                    content_data: str = await response.text()
                    raise ZendirException(
                        f"Request failed with status {response.status}: {content_data}"
                    )

                # Read the response body
                body: bytes = await response.read()

                # Print the response details for debugging
                if body != None and len(body) > 0:
                    response_text = body.decode()
                    if response_text != "null":
                        printer.log(f"Response from {method} {url}: {response_text}")

                # Attempt to decode the response based on the Content-Type header
                if "Content-Type" in response.headers:
                    try:
                        match response.headers["Content-Type"]:
                            case "text/plain":
                                return body.decode()
                            case "application/json":
                                return json.loads(body.decode())
                            case _:
                                return None

                    # Raise an exception if the response cannot be decoded
                    except Exception as e:
                        raise ZendirException(f"Failed to decode response: {e}")

                # Return None if no Content-Type header is present
                return None

    async def get(self, endpoint: str, id: str = "default"):
        """
        Perform an async GET request to the specified endpoint. This will
        return the result of the request as a dictionary.

        :param endpoint: The endpoint to use for the request.
        :type endpoint: str
        :param data: The data to send with the request.
        :type data: Optional[dict]
        :param id: The ID of the context for the client, if applicable.
        :type id: str

        :return: The result of the request.
        :rtype: dict
        """
        return await self._request("GET", endpoint)

    async def post(
        self, endpoint: str, data: Optional[Any] = None, id: str = "default"
    ):
        """
        Perform an async POST request to the specified endpoint. This will
        return the result of the request as a dictionary.

        :param endpoint: The endpoint to use for the request.
        :type endpoint: str
        :param data: The data to send with the request.
        :type data: Optional[dict]
        :param id: The ID of the context for the client, if applicable.
        :type id: str

        :return: The result of the request.
        :rtype: dict
        """
        return await self._request("POST", endpoint, data)

    async def delete(self, endpoint: str, id: str = "default"):
        """
        Perform an async DELETE request to the specified endpoint. This will
        return the result of the request as a dictionary.

        :param endpoint: The endpoint to use for the request.
        :type endpoint: str
        :param data: The data to send with the request.
        :type data: Optional[dict]
        :param id: The ID of the context for the client, if applicable.
        :type id: str

        :return: The result of the request.
        :rtype: dict
        """
        return await self._request("DELETE", endpoint)

    @classmethod
    def create_local(cls, port: int = 25565, timeout: float = 30.0) -> "Client":
        """
        Create a local client with the specified port and request limits.

        :param port: The port to use for the local client.
        :type port: int
        :param timeout: The timeout for requests in seconds.
        :type timeout: float

        :return: A new local client.
        :rtype: Client
        """
        return Client(url=f"http://127.0.0.1:{port}", timeout=timeout)

    def __get_session_info(self) -> List[dict]:
        """
        Returns a list of all session information for the client, assuming a particular
        token is provided. This will return a list of dictionaries, each containing
        information about a session, including the 'guid' of the session, the 'status' of
        the session and the 'version' of the session.

        :return: A list of dictionaries containing session information.
        :rtype: List[dict]
        """

        # Check if the token is provided
        if self.token is None or self.token == "":
            raise ValueError("Token must be provided to list sessions.")

        # Make a GET request to the API endpoint for listing sessions, using the requests library
        printer.log(f"Requesting session information from {self.base_url}.")
        response = requests.get(
            f"{self.base_url}", headers={"X-Api-Key": self.token}, timeout=10
        )
        printer.log(f"Response status code: {response.status_code}")

        # Check if the response was successful
        if response.status_code != 200:
            raise ZendirException(
                f"Failed to list sessions with error: {response.status_code} {response.text}"
            )

        # Assume the response is a JSON array
        try:
            session_info = response.json()
        except json.JSONDecodeError:
            raise ZendirException("Failed to decode response from server.")

        # Ensure the response is a list of session IDs
        if not isinstance(session_info, list):
            raise ZendirException(
                "Failed to list sessions as the server response was not a list."
            )

        # Return the session information
        return session_info

    def __wait_for_session(self, session_id: str, timeout: int = 300) -> bool:
        """
        Waits for a session to become active. This will block until the session is active
        or the timeout is reached. The session will be considered active if the 'status' is
        'RUNNING'. It will make a request every 3 seconds to check the status of the session.

        :param session_id: The ID of the session to wait for.
        :type session_id: str
        :param timeout: The maximum time to wait for the session to become active, in seconds.
        :type timeout: int

        :return: True if the session becomes active, False if the timeout is reached.
        :rtype: bool
        """

        # If there is no session information yet, get it
        if self.session_info is None:
            self.session_info = self.__get_session_info()

        # Adds a flag to print the first time
        first_print: bool = True

        # Get the start time
        start_time = time.time()
        while True:

            # Check if the session is in the session information and is active
            for session in self.session_info:
                if session["guid"] == session_id and session["status"] == "RUNNING":
                    return True

            # If the session is not active, check if the timeout has been reached
            if time.time() - start_time > timeout:
                raise ZendirException(
                    f"Session {session_id} did not become active within {timeout} seconds."
                )

            # If this is the first time, print the message
            if first_print:
                printer.warning(
                    f"Waiting for session '{session_id}' to become active. This may take a few seconds."
                )
                first_print = False

            # Fetch the session information again
            self.session_info = self.__get_session_info()

            # Sleep for 3 seconds before checking again
            time.sleep(3)

    def get_version(self) -> str:
        """
        Fetch the version of the zendir package. This will return the version as a string.

        :return: The version of the zendir package.
        :rtype: str
        """

        # Fetch the version of 'zendir' to ensure compatibility
        version: str = ".".join(zendir_version.split(".")[:3])
        if len(version.split(".")) == 2:
            version += ".0"
        return version

    def create_session(self) -> str:
        """
        Create a new session and return its ID. This will make a POST request to the API
        endpoint for creating a session. The session will be created with the provided token.

        :return: The ID of the newly created session.
        :rtype: str
        """

        # Check if the token is provided
        if self.token is None or self.token == "":
            raise ValueError("Token must be provided to create a session.")

        # Fetch the version of 'zendir' to ensure compatibility
        version: str = self.get_version()

        # Create the data for the POST, including the version property
        data: dict = {
            "version": version,
        }

        # Print a warning that the session is being created
        printer.info(f"Creating a new session with version '{version}'.")

        # Make a POST request to the API endpoint for creating a session
        printer.log(
            f"Requesting session creation at {self.base_url} with data: {data}."
        )
        response = requests.post(
            f"{self.base_url}", headers={"X-Api-Key": self.token}, timeout=60, json=data
        )
        printer.log(f"Response status code: {response.status_code}")

        # Check if the response was successful
        if response.status_code != 200:
            raise ZendirException(
                f"Failed to create session: {response.status_code} {response.text}"
            )

        # Assume the response is a dictionary
        try:
            session_info = response.json()
        except json.JSONDecodeError:
            raise ZendirException(
                "Failed to decode response from server when creating a session."
            )

        # If there is no 'guid' in the response, raise an exception
        if "guid" not in session_info:
            raise ZendirException(
                "Failed to create session: 'guid' not found in response."
            )

        # Return the session ID
        return session_info["guid"]

    def delete_session(self, session_id: str) -> None:
        """
        Delete a session with the given session ID. This will make a DELETE request to the API
        endpoint for deleting a session.

        :param session_id: The ID of the session to delete.
        :type session_id: str
        """

        # Check if the token is provided
        if self.token is None or self.token == "":
            raise ValueError("Token must be provided to delete a session.")

        # Make a DELETE request to the API endpoint for deleting a session
        printer.log(f"Requesting session deletion at {self.base_url}{session_id}/.")
        response = requests.delete(
            f"{self.base_url}{session_id}/",
            headers={"X-Api-Key": self.token},
            timeout=10,
        )
        printer.log(f"Response status code: {response.status_code}")

        # Check if the response was successful
        if response.status_code != 200:
            raise ZendirException(
                f"Failed to delete session {session_id}: {response.status_code} {response.text}"
            )

        # Delete the session from the session information
        printer.success(f"Session '{session_id}' deleted successfully.")

    def list_sessions(self) -> List[str]:
        """
        List all active sessions for the given token. This will return
        a list of session IDs.

        :return: A list of session IDs.
        :rtype: List[str]
        """

        # Get the session information
        self.session_info: List[dict] = self.__get_session_info()

        # Loop through all the sessions and extract the session IDs
        session_list: List[str] = []
        for session in self.session_info:

            # Get the session ID from the session information
            guid: str = session["guid"]
            session_list.append(guid)

        # Return the list of session IDs
        return session_list

    def get_chunk_size(self) -> int:
        """
        Get the maximum chunk size for the client. This is used to limit the size of
        requests to the API.

        :return: The maximum chunk size.
        :rtype: int
        """
        if "127.0.0.1" in self.url or "localhost" in self.url:
            return 1024 * 1024  # 1 MB for local API
        return 6000  # 16 KB for cloud API



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from __future__ import annotations
from ..utils import printer, ZendirException, helper
from .instance import Instance
from .behaviour import Behaviour
from .model import Model
from .message import Message
from .context import Context


class Object(Instance):
    """
    The Object class is able to define an instance that can exist within the simulation.
    An object will have a 3D representation within the simulation and can have behaviours
    and models attached to it. Objects can also have children objects attached to them.
    Objects will always have a position and rotation within the simulation and are the main
    structure for simulation object.
    """

    __instances: dict[str:Instance] = {}
    """Defines all instances that have been connected to the object, by ID."""

    __children: list[Object] = []
    """Defines all children objects that are attached to the object."""

    __behaviours: list[Behaviour] = []
    """Defines all behaviours that are attached to the object."""

    __models: dict[str:Model] = {}
    """Defines all models that are attached to the object, by type."""

    __messages: dict[str:Message] = {}
    """Defines all messages that are attached to the object, by name."""

    __parent: Object = None
    """Defines the parent object that the object is attached to."""

    def __init__(
        self, context: Context, id: str, type: str = None, parent: Object = None
    ) -> None:
        """
        Initialises the object with the context and the ID of the object.

        :param context:         The context of the object
        :type context:          Context
        :param id:              The GUID ID of the object
        :type id:               str
        :param type:            The type of the system, if applicable
        :type type:             str
        :param parent:          The parent object that the object is attached to, if applicable
        :type parent:           Object
        """

        super().__init__(context, id, type)

        # Clear and reset the data
        self.__instances = {}
        self.__children = []
        self.__behaviours = []
        self.__models = {}
        self.__messages = {}
        self.__parent = parent

    @classmethod
    def from_instance(cls, instance: Instance) -> Object:
        """
        Converts the instance object to an object object.

        :param instance:    The instance object to convert
        :type instance:     Instance

        :returns:           The object object that was created
        :rtype:             Object
        """

        # Create the object and set the data
        object = Object(
            instance._context, instance.get_id(), instance._Instance__type, None
        )
        object.__dict__ = instance.__dict__
        object._Instance__data = instance._Instance__data
        object._refresh_cache = instance._refresh_cache

        # Return the object
        return object

    async def _get_data(self) -> None:
        """
        Overrides the base class method to fetch the data from the API and store it in the
        object. This is used to ensure that the data is fetched correctly and is up to date.
        Additionally, this function will also fetch all the children, behaviours, models and
        messages that are attached to the object.
        """

        # Fetch the base data
        if not self._refresh_cache:
            return
        self._ignore_refresh_override = True
        await super()._get_data()

        # Reload the heirarchies of the data
        await self.__reload_heirarchy(recurse=False)

        # Now, set the override and the required refresh again
        self._ignore_refresh_override = False
        self._refresh_cache = self._context.always_require_refresh

    async def __reload_heirarchy(self, recurse: bool = True) -> None:
        """
        Reloads the heirarchy of the object by fetching all the children, behaviours,
        and models that are attached to the object. This will also ensure that all
        objects are registered with the object and that the data is up to date.

        :param recurse:  Whether to recursively reload the heirarchy of the children
        :type recurse:   bool
        """

        # Get the children of the object
        children_ids: list[str] = await self.get("Children")
        for child_id in children_ids:
            if child_id not in self.__instances:
                self.__register_child(child_id, type="")

        # Get the behaviours of the object
        behaviours_ids: list[str] = await self.get("Behaviours")
        for behaviour_id in behaviours_ids:
            if behaviour_id not in self.__instances:
                self.__register_behaviour(behaviour_id, type="")

        # Now, get the models of the object
        models_ids: list[str] = await self.get("Models")
        for model_id in models_ids:
            if model_id not in self.__instances:
                model = Model(self._context, model_id, None, parent=self)
                self.__instances[model_id] = model
                self.__models[await model.get_type()] = model

        # Now, for each of the children, reload the heirarchy
        if recurse:
            for child in self.__children:
                await child.__reload_heirarchy()

    def _reset_refresh_cache(self) -> None:
        """
        Overrides the base class method to set the flag for refreshing the cache to true.
        This will ensure that all sub-objects will also require a refresh.
        """

        # Ensure all sub-objects require a refresh too
        for _, instance in self.__instances.items():
            instance._reset_refresh_cache()
        super()._reset_refresh_cache()

    def get_parent(self) -> Object:
        """
        Returns the parent object that the object is attached to, if it exists.

        :returns:   The parent object that the object is attached to
        :rtype:     Object
        """

        return self.__parent

    def get_instance_with_id(self, id: str, recurse: bool = False) -> Instance:
        """
        Returns the instance that is attached to the object with the specified ID. If the
        instance does not exist, None will be returned.

        :param id:  The ID of the instance to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of children to find the instance
        :type recurse:   bool

        :returns:   The instance that is attached to the object with the specified ID
        :rtype:     Instance
        """

        # Start by looking at the instances for the ID
        if id in self.__instances:
            return self.__instances[id]

        # If recurse is enabled, look down the chain of children
        if recurse:
            for child in self.__children:
                result = child.get_instance_with_id(id, recurse)
                if result:
                    return result
            for behaviour in self.__behaviours:
                result = behaviour.get_instance_with_id(id)
                if result:
                    return result
            for model in self.__models.values():
                result = model.get_instance_with_id(id)
                if result:
                    return result

        # Return None if the instance is not found
        return None

    async def add_child(self, type: str, **kwargs) -> Object:
        """
        Adds a child object to the object with the specified type. The child object will
        be created and attached to the object and will be returned to the user.

        :param type:    The type of the child object to create
        :type type:     str
        :param kwargs:  Optional additional data to set on the child object when created
        :type kwargs:   dict

        :returns:       The child object that was created
        :rtype:         Object
        """

        # Check the type and validate it
        type = helper.validate_type(type)

        # Invoke the function library to create the object
        child_id: str = await self.invoke("AddObject", type)

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(child_id):
            raise ZendirException(f"Failed to create child object of type '{type}'.")

        # Create the object and add it to the array
        child: Object = self.__register_child(child_id, type)
        if child is None:
            raise ZendirException(f"Failed to create child object of type '{type}'.")

        # If there are any kwargs, set them on the child object
        if len(kwargs) > 0:
            await child.set(**kwargs)

        # Regsiter the child object with the ID
        return child

    def __register_child(self, id: str, type: str = "") -> Object:
        """
        Registers a child object to the object with the specified ID. The child object will
        be created and attached to the object and will be returned to the user.

        :param id:  The ID of the child object to create
        :type id:   str
        :param type:    The type of the child object to create
        :type type:     str

        :returns:   The child object that was created
        :rtype:     Object
        """

        # Create the object
        object = Object(self._context, id, type, parent=self)
        self.__children.append(object)
        self.__instances[id] = object

        # Print the success message
        if type != "":
            printer.success(f"Successfully created child object of type '{type}'.")
        return object

    def __is_child_registered(self, id: str) -> bool:
        """
        Checks if a child object with the specified ID is registered. This will
        check down the hierarchy of child objects.

        :param id:  The ID of the child object to check
        :type id:   str

        :returns:   True if the child object is registered, False otherwise
        :rtype:     bool
        """

        # Check if the ID is in the instances
        if id in self.__instances:
            return True

        # Check down the hierarchy of child objects
        for child in self.__children:
            if child.__is_child_registered(id):
                return True

        # If not found, return False
        return False

    def __get_registered_child(self, id: str, recurse: bool = True) -> Object:
        """
        Returns the child object that is attached to the object with the specified ID. If the
        child object does not exist, None will be returned. This will also look down the chain
        of instances to find the child object, if specified.

        :param id:  The ID of the child object to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of instances to find the child object
        :type recurse:   bool

        :returns:   The child object that is attached to the object with the specified ID
        :rtype:     Object
        """

        # Start by looking at the children for the ID
        for child in self.__children:
            if child.get_id() == id:
                return child

        # If recurse is enabled, look down the chain of instances
        if recurse:
            for child in self.__children:
                result = child.__get_registered_child(id, recurse)
                if result:
                    return result

        # Return None if the child object is not found
        return None

    def get_child(self, index: int) -> Object:
        """
        Returns the child object at the specified index. If the index is invalid, an
        exception will be raised.

        :param index:   The index of the child object to fetch
        :type index:    int

        :returns:       The child object at the specified index
        :rtype:         Object
        """

        # Fetch the child and perform a safety check
        if index < 0 or index >= len(self.__children):
            raise IndexError(f"Failed to get child object at index: {index}.")
        return self.__children[index]

    def get_children(self) -> list[Object]:
        """
        Returns all of the children objects that are attached to the object.

        :returns:   All of the children objects that are attached to the object
        :rtype:     list[Object]
        """

        return self.__children

    async def find_child_with_type(self, type: str, recurse: bool = True) -> Object:
        """
        Finds the first child object that is attached to the object of the specified
        type. If the type is not found, None will be returned.

        :param type:    The type of the child object to fetch
        :type type:     str
        :param recurse: Whether to search recursively through child objects
        :type recurse:  bool

        :returns:       The first child object that is attached to the object of the specified type
        :rtype:         Object
        """

        # Fetch the children with the specified type and return the first one
        children: list[Object] = await self.find_children_with_type(
            type, recurse=recurse
        )
        if len(children) > 0:
            return children[0]

        # If no children were found, return None
        return None

    async def find_children_with_type(
        self, type: str, recurse: bool = True
    ) -> list[Object]:
        """
        Finds all of the children objects that are attached to the object of the specified
        type. If the type is not found, an empty list will be returned.

        :param type:    The type of the children objects to fetch
        :type type:     str
        :param recurse: Whether to search recursively through child objects
        :type recurse:  bool

        :returns:       All of the children objects that are attached to the object of the specified type
        :rtype:         list[Object]
        """

        # Check the type and validate it
        type = helper.validate_type(type)

        # Fetch the children with the specified type
        children_ids: list[str] = await self.invoke(
            "FindChildrenWithType", type, recurse
        )

        # For each child, check if it exists. If it does not, we need to require
        # a reload of the heirarchy to ensure that the child is registered.
        require_reload: bool = False
        for child_id in children_ids:
            if not self.__is_child_registered(child_id):
                require_reload = True
                break

        # If required a reload, reload the heirarchy
        if require_reload:
            await self.__reload_heirarchy(recurse=True)

        # Now, create an array with all children from the IDs
        children: list[Object] = [
            self.__get_registered_child(child_id, recurse=True)
            for child_id in children_ids
        ]

        # Return the list
        return children

    async def find_child_with_id(self, id: str, recurse: bool = True) -> Object:
        """
        Returns the child object that is attached to the object with the specified ID. If the
        child object does not exist, None will be returned. This will also look down the chain
        of instances to find the child object, if specified.

        :param id:  The ID of the child object to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of instances to find the child object
        :type recurse:   bool

        :returns:   The child object that is attached to the object with the specified ID
        :rtype:     Object
        """

        # Fetch the child with the specified ID
        child_id: str = await self.invoke("FindChildWithID", id, recurse)

        # If the child ID is not valid, return None
        if not helper.is_valid_guid(child_id):
            return None

        # Check if the child is not registered, and reload the heirarchy
        if not self.__is_child_registered(child_id):
            await self.__reload_heirarchy(recurse=True)

        # Now, return the registered child with the ID
        child: Object = self.__get_registered_child(child_id, recurse=recurse)
        if child is None:
            return None

        # If the child is found, return it
        return child

    async def add_behaviour(self, type: str, **kwargs) -> Behaviour:
        """
        Adds a behaviour to the object with the specified type. The behaviour will be created
        and attached to the object and will be returned to the user.

        :param type:    The type of the behaviour to create
        :type type:     str
        :param kwargs:  Optional additional data to set on the behaviour when created
        :type kwargs:   dict

        :returns:       The behaviour that was created
        :rtype:         Behaviour
        """

        # Check the type and validate it
        type = helper.validate_type(type)

        # Invoke the function library to create the behaviour
        behaviour_id: str = await self.invoke("AddObject", type)

        # If the ID is not valid, raise an exception
        if not helper.is_valid_guid(behaviour_id):
            raise ZendirException(f"Failed to create child behaviour of type '{type}'.")

        # Create the behaviour and add it to the array
        behaviour: Behaviour = self.__register_behaviour(behaviour_id, type)
        if behaviour is None:
            raise ZendirException(f"Failed to create child behaviour of type '{type}'.")

        # If there are any kwargs, set them on the child behaviour
        if len(kwargs) > 0:
            await behaviour.set(**kwargs)

        # Regsiter the child behaviour with the ID
        return behaviour

    def __register_behaviour(self, id: str, type: str = "") -> Behaviour:
        """
        Registers a child behaviour to the object with the specified ID. The child behaviour will
        be created and attached to the behaviour and will be returned to the user.

        :param id:  The ID of the child behaviour to create
        :type id:   str
        :param type:    The type of the child behaviour to create
        :type type:     str

        :returns:   The child behaviour that was created
        :rtype:     Object
        """

        # Create the behaviour
        behaviour = Behaviour(self._context, id, type, parent=self)
        self.__behaviours.append(behaviour)
        self.__instances[id] = behaviour

        # Print the success message
        if type != "":
            printer.success(f"Successfully created child behaviour of type '{type}'.")
        return behaviour

    def __is_behaviour_registered(self, id: str) -> bool:
        """
        Checks if a child behaviour with the specified ID is registered. This will
        check down the hierarchy of child behaviours.

        :param id:  The ID of the child behaviour to check
        :type id:   str

        :returns:   True if the child behaviour is registered, False otherwise
        :rtype:     bool
        """

        # Check if the ID is in the instances
        if id in self.__instances:
            return True

        # Check down the hierarchy of child objects
        for child in self.__children:
            if child.__is_behaviour_registered(id):
                return True

        # If not found, return False
        return False

    def __get_registered_behaviour(self, id: str, recurse: bool = True) -> Behaviour:
        """
        Returns the behaviour that is attached to the object with the specified ID. If the
        behaviour does not exist, None will be returned. This will also look down the chain
        of instances to find the behaviour, if specified.

        :param id:  The ID of the behaviour to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of instances to find the child object
        :type recurse:   bool

        :returns:   The behaviour that is attached to the object with the specified ID
        :rtype:     Behaviour
        """

        # Start by looking at the children for the ID
        for behaviour in self.__behaviours:
            if behaviour.get_id() == id:
                return behaviour

        # If recurse is enabled, look down the chain of instances
        if recurse:
            for child in self.__children:
                result = child.__get_registered_behaviour(id, recurse)
                if result:
                    return result

        # Return None if the behaviour is not found
        return None

    def get_behaviour(self, index: int) -> Behaviour:
        """
        Gets the behaviour at the specified index. If the index is invalid, an exception
        will be raised.

        :param index:   The index of the behaviour to fetch
        :type index:    int

        :returns:       The behaviour at the specified index
        :rtype:         Behaviour
        """

        # Fetch the child and perform a safety check
        if index < 0 or index >= len(self.__behaviours):
            raise IndexError(f"Failed to get behaviour at index: {index}.")
        return self.__behaviours[index]

    def get_behaviours(self) -> list[Behaviour]:
        """
        Returns all of the behaviours that are attached to the object.

        :returns:   All of the behaviours that are attached to the object
        :rtype:     list[Behaviour]
        """

        return self.__behaviours

    async def find_behaviour_with_type(
        self, type: str, recurse: bool = True
    ) -> Behaviour:
        """
        Finds the first behaviour that is attached to the object of the specified
        type. If the type is not found, None will be returned.

        :param type:    The type of the behaviour to fetch
        :type type:     str
        :param recurse: Whether to search recursively through child objects
        :type recurse:  bool

        :returns:       The first behaviour that is attached to the object of the specified type
        :rtype:         Behaviour
        """

        # Fetch the children with the specified type and return the first one
        behaviours: list[Behaviour] = await self.find_behaviours_with_type(
            type, recurse=recurse
        )
        if len(behaviours) > 0:
            return behaviours[0]

        # If no behaviours were found, return None
        return None

    async def find_behaviours_with_type(
        self, type: str, recurse: bool = True
    ) -> list[Behaviour]:
        """
        Finds all of the behaviours that are attached to the object of the specified
        type. If the type is not found, an empty list will be returned.

        :param type:    The type of the behaviours to fetch
        :type type:     str
        :param recurse: Whether to search recursively through child objects
        :type recurse:  bool

        :returns:       All of the behaviours that are attached to the object of the specified type
        :rtype:         list[Behaviour]
        """

        # Check the type and validate it
        type = helper.validate_type(type)

        # Fetch the behaviours with the specified type
        behaviours_ids: list[str] = await self.invoke(
            "FindBehavioursWithType", type, recurse
        )

        # For each behaviour, check if it exists. If it does not, we need to require
        # a reload of the heirarchy to ensure that the behaviour is registered.
        require_reload: bool = False
        for behaviour_id in behaviours_ids:
            if not self.__is_behaviour_registered(behaviour_id):
                require_reload = True
                break

        # If required a reload, reload the heirarchy
        if require_reload:
            await self.__reload_heirarchy(recurse=True)

        # Now, create an array with all behaviours from the IDs
        behaviours: list[Behaviour] = [
            self.__get_registered_behaviour(behaviour_id, recurse=True)
            for behaviour_id in behaviours_ids
        ]

        # Return the list
        return behaviours

    async def find_behaviour_with_id(self, id: str, recurse: bool = True) -> Behaviour:
        """
        Returns the behaviour that is attached to the object with the specified ID. If the
        behaviour does not exist, None will be returned. This will also look down the chain
        of instances to find the behaviour, if specified.

        :param id:  The ID of the behaviour to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of instances to find the behaviour
        :type recurse:   bool

        :returns:   The behaviour that is attached to the object with the specified ID
        :rtype:     Behaviour
        """

        # Fetch the behaviour with the specified ID
        behaviour_id: str = await self.invoke("FindBehaviourWithID", id, recurse)

        # If the behaviour ID is not valid, return None
        if not helper.is_valid_guid(behaviour_id):
            return None

        # Check if the behaviour is not registered, and reload the heirarchy
        if not self.__is_behaviour_registered(behaviour_id):
            await self.__reload_heirarchy(recurse=True)

        # Now, return the registered behaviour with the ID
        behaviour: Behaviour = self.__get_registered_behaviour(
            behaviour_id, recurse=recurse
        )
        if behaviour is None:
            return None

        # If the behaviour is found, return it
        return behaviour

    async def get_model(self, type: str, **kwargs) -> Model:
        """
        Attempts to get the model of the specified type that is attached to the object. If the
        model does not exist, it will be created and attached to the object. If the model cannot
        be created, an exception will be raised.

        :param type:    The type of the model to fetch
        :type type:     str
        :param kwargs:  The additional data to add to the model
        :type kwargs:   dict

        :returns:       The model of the specified type that is attached to the object
        :rtype:         Model
        """

        # Check the type and validate it
        type = helper.validate_type(type)

        # For each of the kwargs, serialize the data
        for key in kwargs:
            kwargs[key] = helper.serialize(kwargs[key])

        # Check to see if the model exists
        if type in self.__models.keys():
            model: Model = self.__models[type]
            if len(kwargs) > 0:
                await model.set(**kwargs)
            return model

        # Attempt to find or create the model
        id: str = await self._context.get_client().post(
            f"{self.get_id()}/ivk", ["GetModel", type], id=self._context.get_id()
        )
        if not helper.is_valid_guid(id):
            raise ZendirException(f"Failed to create model of type '{type}'.")

        # Create the model with the ID
        model = self.__register_model(id, type)

        # Set the data if it exists
        if len(kwargs) > 0:
            await model.set(**kwargs)
        return model

    def __register_model(self, id: str, type: str = "") -> Model:
        """
        Registers a model to the object with the specified ID. The modelr will
        be created and attached to the object and will be returned to the user.

        :param id:      The ID of the model to create
        :type id:       str
        :param type:    The type of the model to create
        :type type:     str

        :returns:   The model that was created
        :rtype:     Object
        """

        # Create the model with the ID
        model = Model(self._context, id, type, parent=self)
        self.__models[type] = model
        self.__instances[id] = model

        # Print the success message
        printer.success(f"Successfully created model of type '{type}'.")
        return model

    def get_models(self) -> list[Model]:
        """
        Returns all of the models that are attached to the object.

        :returns:   All of the models that are attached to the object
        :rtype:     list[Model]
        """

        return list(self.__models.values())

    async def get_message(self, name: str) -> Message:
        """
        Attempts to get the message with the specified name that is attached to the object. If the
        message does not exist, it will be created and attached to the object. If the message cannot
        be created, an exception will be raised.

        :param name:    The name of the message to fetch
        :type name:     str

        :returns:       The message with the specified name that is attached to the object
        :rtype:         Message
        """

        # Check if the name is within the message structure and return that
        if name in self.__messages.keys():
            return self.__messages[name]

        # Fetch the data
        message_id: str = await self.get(name)
        if not helper.is_valid_guid(message_id):
            raise ZendirException(f"Failed to find message with name '{name}'.")

        # Create the message object with the ID
        message = Message(self._context, message_id)
        self.__messages[name] = message
        self.__instances[message_id] = message

        # Return the message of that name
        printer.success(f"Successfully created message with name '{name}'.")
        return message

    async def get_messages(self) -> list[Message]:
        """
        Returns all of the messages that are attached to the object. This will only include the
        messages that have currently been fetched.

        :returns:   All of the messages that are attached to the object
        :rtype:     list[Message]
        """

        # Fetch all values on the object
        data: dict = await self.get_all()

        # If any data starts with 'Out_', then it is a message
        for key in data.keys():
            if str(key).startswith("Out_"):
                await self.get_message(key)
            if str(key).startswith("In_"):
                if helper.is_valid_guid(data[key]):
                    await self.get_message(key)

        # Return all the messages
        return self.__messages.values()



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

"""
This module contains public methods that can calculate various
data-related functions for the TT&C system. These include bits
and byte conversions.
"""


def bytes_to_bits(bytes: float) -> int:
    """
    Convert a quantity from bytes to bits. This function calculates
    the number of bits that correspond to a given number of bytes.
    Note that 1 byte equals 8 bits.

    :param bytes:       The number of bytes to be converted.
    :type bytes:        float

    :return:            The equivalent number of bits.
    :rtype: int
    """
    if bytes < 0:
        raise ValueError("bytes must be a positive float")
    try:
        bytes = float(bytes)
    except TypeError:
        raise TypeError("bytes must be a positive float")
    return int(bytes * 8)


def kilobytes_to_bytes(kilobytes: float) -> int:
    """
    Convert a quantity from kilobytes to bytes. This function
    uses the `bytes_to_bytes` function for the conversion, considering
    that 1 kilobyte equals 1024 bytes.

    :param kilobytes:   The number of kilobytes to be converted.
    :type kilobytes:    float

    :return:            The equivalent number of bits.
    :rtype:             int
    """
    if kilobytes < 0:
        raise ValueError("kilobytes must be a positive float")
    try:
        kilobytes = float(kilobytes)
    except TypeError:
        raise TypeError("kilobytes must be a positive float")
    return int(kilobytes * 1024)


def megabytes_to_bytes(megabytes: float) -> int:
    """
    Convert a quantity from megabytes to bytes. This function uses the
    `kilobytes_to_bytes` function for the conversion, considering
    that 1 megabyte equals 1024 kilobytes

    :param megabytes:   The number of megabytes to be converted.
    :type megabytes:    float

    :return:            The equivalent number of bits.
    :rtype:             int
    """
    return kilobytes_to_bytes(megabytes * 1024)


def gigabytes_to_bytes(gigabytes: float) -> int:
    """
    Convert a quantity from gigabytes to bytes. This function uses
    the `megabytes_to_bytes` function for the conversion, considering
    that 1 gigabyte equals 1024 megabytes.

    :param gigabytes:   The number of gigabytes to be converted.
    :type gigabytes:    float

    :return:            The equivalent number of bits.
    :rtype:             int
    """
    return megabytes_to_bytes(gigabytes * 1024)


def kilobytes_to_bits(kilobytes: float) -> int:
    """
    Convert a quantity from kilobytes to bits.  This function
    uses the `bytes_to_bits` function for the conversion, considering
    that 1 kilobyte equals 1024 bytes.

    :param kilobytes:   The number of kilobytes to be converted.
    :type kilobytes:    float

    :return:            The equivalent number of bits.
    :rtype:             int
    """
    return bytes_to_bits(kilobytes_to_bytes(kilobytes))


def megabytes_to_bits(megabytes: float) -> int:
    """
    Convert a quantity from megabytes to bits.

    :param megabytes:   The number of megabytes to be converted.
    :type megabytes:    float

    :return:            The equivalent number of bits.
    :rtype:             int
    """
    return bytes_to_bits(megabytes_to_bytes(megabytes))


def gigabytes_to_bits(gigabytes: float) -> int:
    """
    Convert a quantity from gigabytes to bits.

    :param gigabytes:   The number of gigabytes to be converted.
    :type gigabytes:    float

    :return:            The equivalent number of bits.
    :rtype:             int
    """
    return bytes_to_bits(gigabytes_to_bytes(gigabytes))



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.
import re
from pathlib import Path
from setuptools import setup, find_packages


def get_version():
    init_file = Path(__file__).parent / "src" / "zendir" / "__init__.py"
    content = init_file.read_text()
    match = re.search(r'__version__\s*=\s*["\']([^"\']+)["\']', content)
    if match:
        return match.group(1)
    raise RuntimeError("Cannot find version string.")


# Setup the project
setup(
    name="zendir",
    version=get_version(),
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "aiohttp",
        "urllib3",
        "paho-mqtt",
        "numpy",
        "pandas",
        "matplotlib",
        "setuptools",
        "pytest-asyncio",
        "requests",
    ],
    author="Zendir",
    author_email="support@zendir.io",
    description="Python Interface to the Zendir API for simulations",
    long_description="This is the Python interface library for the Zendir API. \
        The Zendir API allows access to spacecraft and space-domain simulation functions \
        and the framework for simulating high-fidelity satellites. \
        It enables accessing the REST API simulation functions in an easy format.\
        Examples of how to construct Zendir simulations are provided on the\
        Zendir documentation found at https://docs.zendir.io.",
    url="https://github.com/zendir-dev/zendir-py",
)
